import random
import string

from behave import *
from claudia.python.lib.framework import constants
from claudia.python.lib.framework.common import *
from claudia.python.lib.framework.object_factory import ObjFactory, ObjType
from hamcrest import assert_that, equal_to, contains_string
from xrpl.account import get_balance
from xrpl.models import AccountSet, AccountSetAsfFlag, Tx


@when('we mint NFT using same token taxon')
def step_impl(context):
    try:
        payload = ObjFactory.getObj(
            ObjType.nft_token_mint,
            account=context.sourceAddress,
            nftoken_taxon=0,
        )
        context.nft_mint_response_1 = sign_autofill_and_submit(context.client, payload, context.sourceWallet)
        context.nft_mint_response_2 = sign_autofill_and_submit(context.client, payload, context.sourceWallet)
    except Exception as e:
        context.exception = e
        context.test_status = "failed"

        raise e


@then('nft minting is successful')
def step_impl(context):
    verify_account_balance(get_balance(context.sourceAddress, context.client),
                           (int(context.sourceBalance) - (2 * context.minDrop)))
    verify_nftoken_mint_account_objects(context, context.nft_mint_response_1)
    verify_nftoken_mint_account_objects(context, context.nft_mint_response_2)


@when('we mint NFT with low reserves')
def step_impl(context):
    try:
        payload = ObjFactory.getObj(
            ObjType.nft_token_mint,
            account=context.sourceAddress,
            nftoken_taxon=0,
        )
        context.nft_mint_response_1 = sign_autofill_and_submit(context.client, payload, context.sourceWallet)
    except Exception as e:
        context.exception = e
        context.test_status = "failed"


@then('nft minting with low reserves is not successful')
def step_impl(context):
    verify_test_status(context.test_status, "failed")
    verify_test_error_message(str(context.exception), "Transaction failed: tecINSUFFICIENT_RESERVE")
    verify_account_balance(get_balance(context.sourceAddress, context.client), context.sourceBalance - context.minDrop)


@when('we mint NFT with optional URI as a string')
def step_impl(context):
    try:
        payload = ObjFactory.getObj(
            ObjType.nft_token_mint,
            account=context.sourceAddress,
            nftoken_taxon=0,
            uri="ipfs://bafybeigdyrzt5sfp7udm7hu76uh7y26nf3efuylqabf3oclgtqy55fbzdi"
        )
        context.nft_mint_response_1 = sign_autofill_and_submit(context.client, payload, context.sourceWallet)
    except Exception as e:
        context.exception = e
        context.test_status = "failed"


@then('nft minting with optional URI as a string is not successful')
def step_impl(context):
    verify_test_status(context.test_status, "failed")
    verify_test_error_message(str(context.exception), "non-hexadecimal number found")
    verify_account_balance(get_balance(context.sourceAddress, context.client), int(context.sourceBalance))


@when('we mint NFT with optional URI as a hex string')
def step_impl(context):
    try:
        uri = "ipfs://bafybeigdyrzt5sfp7udm7hu76uh7y26nf3efuylqabf3oclgtqy55fbzdi"
        hex_uri = uri.encode('utf-8').hex()
        payload = ObjFactory.getObj(
            ObjType.nft_token_mint,
            account=context.sourceAddress,
            nftoken_taxon=0,
            uri=hex_uri
        )
        context.nft_mint_response_1 = sign_autofill_and_submit(context.client, payload, context.sourceWallet)
    except Exception as e:
        context.exception = e
        context.test_status = "failed"
        raise e


@then('nft minting with optional URI as a hex string is not successful')
def step_impl(context):
    verify_account_balance(get_balance(context.sourceAddress, context.client),
                           (int(context.sourceBalance) - (context.minDrop)))
    verify_nftoken_mint_account_objects(context, context.nft_mint_response_1)


@when('we mint NFT with invalid URI')
def step_impl(context):
    try:
        uri = ""
        hex_uri = uri.encode('utf-8').hex()
        payload = ObjFactory.getObj(
            ObjType.nft_token_mint,
            account=context.sourceAddress,
            nftoken_taxon=0,
            uri=hex_uri
        )
        context.nft_mint_response_1 = sign_autofill_and_submit(context.client, payload, context.sourceWallet)
    except Exception as e:
        context.exception = e
        context.test_status = "failed"


@then('nft minting with invalid URI is not successful')
def step_impl(context):
    verify_test_status(context.test_status, "failed")
    verify_test_error_message(str(context.exception), "Malformed transaction")
    verify_account_balance(get_balance(context.sourceAddress, context.client), int(context.sourceBalance))


@when('we mint NFT with optional URI with more than 265 characters')
def step_impl(context):
    try:
        uri_length = 257
        uri = "ipfs://bafybeigdyrzt5sfp7udm7hu76uh7y26nf3efuylqabf3oclgtqy55fbzdi"
        for count in range(uri_length - len(uri)):
            uri += random.choice(string.ascii_letters)

        hex_uri = uri.encode('utf-8').hex()
        payload = ObjFactory.getObj(
            ObjType.nft_token_mint,
            account=context.sourceAddress,
            nftoken_taxon=0,
            uri=hex_uri
        )
        context.nft_mint_response_1 = sign_autofill_and_submit(context.client, payload, context.sourceWallet)
    except Exception as e:
        context.exception = e
        context.test_status = "failed"


@then('nft minting with optional URI with more than 265 characters is not successful')
def step_impl(context):
    verify_test_status(context.test_status, "failed")
    verify_test_error_message(str(context.exception), "Must not be longer than 512 characters")
    verify_account_balance(get_balance(context.sourceAddress, context.client), int(context.sourceBalance))


@when('we mint NFT with optional {transfer_fee} transfer fee')
def step_impl(context, transfer_fee):
    try:
        fee = int(sanitize_args(transfer_fee))
        payload = ObjFactory.getObj(
            ObjType.nft_token_mint,
            account=context.sourceAddress,
            transfer_fee=fee,
            flags=8,
            nftoken_taxon=0
        )
        context.nft_mint_response_1 = sign_autofill_and_submit(context.client, payload, context.sourceWallet)
    except Exception as e:
        context.exception = e
        context.test_status = "failed"


@then('nft minting with optional negative transfer fee is not successful')
def step_impl(context):
    verify_test_status(context.test_status, "failed")
    verify_test_error_message(str(context.exception), "can't convert negative int to unsigned")
    verify_account_balance(get_balance(context.sourceAddress, context.client), int(context.sourceBalance))


@then('nft minting with optional zero transfer fee is successful')
def step_impl(context):
    verify_account_balance(get_balance(context.sourceAddress, context.client),
                           int(context.sourceBalance) - context.minDrop)
    verify_nftoken_mint_account_objects(context, context.nft_mint_response_1)


@then('nft minting with optional transfer fee in decimal is not successful')
def step_impl(context):
    verify_test_status(context.test_status, "failed")
    verify_test_error_message(str(context.exception), "invalid literal for int() with base 10")
    verify_account_balance(get_balance(context.sourceAddress, context.client), int(context.sourceBalance))


@then('nft minting with optional transfer fee more than 50000 is not successful')
def step_impl(context):
    verify_test_status(context.test_status, "failed")
    verify_test_error_message(str(context.exception), "'transfer_fee': 'Must not be greater than 50000'")
    verify_account_balance(get_balance(context.sourceAddress, context.client), int(context.sourceBalance))


@when('we mint NFT without tfTransferable flag')
def step_impl(context):
    try:
        payload = ObjFactory.getObj(
            ObjType.nft_token_mint,
            account=context.sourceAddress,
            transfer_fee=50,
            # flags=8,
            nftoken_taxon=0
        )
        context.nft_mint_response_1 = sign_autofill_and_submit(context.client, payload, context.sourceWallet)
    except Exception as e:
        context.exception = e
        context.test_status = "failed"


@then('nft minting without tfTransferable flag is not successful')
def step_impl(context):
    verify_test_status(context.test_status, "failed")
    verify_test_error_message(str(context.exception), "Malformed transaction")
    verify_account_balance(get_balance(context.sourceAddress, context.client), int(context.sourceBalance))


@when('we mint NFT with optional memos')
def step_impl(context):
    try:
        uri = "ipfs://bafybeigdyrzt5sfp7udm7hu76uh7y26nf3efuylqabf3oclgtqy55fbzdi"
        hex_uri = uri.encode('utf-8').hex()
        payload = ObjFactory.getObj(
            ObjType.nft_token_mint,
            account=context.sourceAddress,
            uri=hex_uri,
            memos=[
                ObjFactory.getObj(
                    ObjType.memo,
                    memo_type="687474703A2F2F6578616D706C652E636F6D2F6D656D6F2F67656E65726963",
                    memo_data="72656E74"
                )
            ],
            nftoken_taxon=0
        )
        context.nft_mint_response_1 = sign_autofill_and_submit(context.client, payload, context.sourceWallet)
    except Exception as e:
        context.exception = e
        context.test_status = "failed"


@then('nft minting with optional memos is successful')
def step_impl(context):
    verify_account_balance(get_balance(context.sourceAddress, context.client),
                           int(context.sourceBalance) - context.minDrop)
    verify_nftoken_mint_account_objects(context, context.nft_mint_response_1)


@when('we mint NFT with bad seed')
def step_impl(context):
    try:
        payload = ObjFactory.getObj(
            ObjType.nft_token_mint,
            account=context.sourceAddress,
            nftoken_taxon=0,
        )
        context.nft_mint_response_1 = sign_autofill_and_submit(context.client, payload, context.destinationWallet)
    except Exception as e:
        context.exception = e
        context.test_status = "failed"


@then('nft minting with bad seed is not successful')
def step_impl(context):
    verify_test_status(context.test_status, "failed")
    verify_test_error_message(str(context.exception), "Prelim result: tefBAD_AUTH")
    verify_account_balance(get_balance(context.sourceAddress, context.client), int(context.sourceBalance))


@when('we mint NFT without nftoken_taxon')
def step_impl(context):
    try:
        payload = ObjFactory.getObj(
            ObjType.nft_token_mint,
            account=context.sourceAddress,
        )
        context.nft_mint_response_1 = sign_autofill_and_submit(context.client, payload, context.sourceWallet)
    except Exception as e:
        context.exception = e
        context.test_status = "failed"


@then('nft minting without nftoken_taxon is not successful')
def step_impl(context):
    verify_test_status(context.test_status, "failed")
    verify_test_error_message(str(context.exception), "nftoken_taxon is not set")
    verify_account_balance(get_balance(context.sourceAddress, context.client), int(context.sourceBalance))


@when('we mint NFT with negative nftoken_taxon')
def step_impl(context):
    try:
        payload = ObjFactory.getObj(
            ObjType.nft_token_mint,
            account=context.sourceAddress,
            nftoken_taxon=-10,
        )
        context.nft_mint_response_1 = sign_autofill_and_submit(context.client, payload, context.sourceWallet)
    except Exception as e:
        context.exception = e
        context.test_status = "failed"


@then('nft minting with negative nftoken_taxon is not successful')
def step_impl(context):
    verify_test_status(context.test_status, "failed")
    verify_test_error_message(str(context.exception), "can't convert negative int to unsigned")
    verify_account_balance(get_balance(context.sourceAddress, context.client), int(context.sourceBalance))


@when('we mint NFT with too high nftoken_taxon')
def step_impl(context):
    try:
        payload = ObjFactory.getObj(
            ObjType.nft_token_mint,
            account=context.sourceAddress,
            nftoken_taxon=4294967296,
        )
        context.nft_mint_response_1 = sign_autofill_and_submit(context.client, payload, context.sourceWallet)
    except Exception as e:
        context.exception = e
        context.test_status = "failed"


@then('nft minting with too high nftoken_taxon is not successful')
def step_impl(context):
    verify_test_status(context.test_status, "failed")
    verify_test_error_message(str(context.exception), "int too big to convert")
    verify_account_balance(get_balance(context.sourceAddress, context.client), int(context.sourceBalance))


@when('we mint NFT without source account field')
def step_impl(context):
    try:
        payload = ObjFactory.getObj(
            ObjType.nft_token_mint,
            nftoken_taxon=0,
        )
        context.nft_mint_response_1 = sign_autofill_and_submit(context.client, payload, context.sourceWallet)
    except Exception as e:
        context.exception = e
        context.test_status = "failed"


@then('nft minting without source account field is not successful')
def step_impl(context):
    verify_test_status(context.test_status, "failed")
    verify_test_error_message(str(context.exception), "account is not set")
    verify_account_balance(get_balance(context.sourceAddress, context.client), int(context.sourceBalance))


@when('we mint NFT with empty source account field')
def step_impl(context):
    try:
        payload = ObjFactory.getObj(
            ObjType.nft_token_mint,
            account="",
            nftoken_taxon=0,
        )
        context.nft_mint_response_1 = sign_autofill_and_submit(context.client, payload, context.sourceWallet)
    except Exception as e:
        context.exception = e
        context.test_status = "failed"


@then('nft minting with empty source account field is not successful')
def step_impl(context):
    verify_test_status(context.test_status, "failed")
    verify_test_error_message(str(context.exception), "Request failed, actMalformed: Account malformed")
    verify_account_balance(get_balance(context.sourceAddress, context.client), int(context.sourceBalance))


@when('we mint NFT with same account as issuer')
def step_impl(context):
    try:
        payload = ObjFactory.getObj(
            ObjType.nft_token_mint,
            account=context.sourceAddress,
            issuer=context.sourceAddress,
            nftoken_taxon=0,
        )
        context.nft_mint_response_1 = sign_autofill_and_submit(context.client, payload, context.sourceWallet)
    except Exception as e:
        context.exception = e
        context.test_status = "failed"


@then('nft minting with same account as issuer is not successful')
def step_impl(context):
    verify_test_status(context.test_status, "failed")
    verify_test_error_message(str(context.exception), "'issuer': 'Must not be the same as the account'")
    verify_account_balance(get_balance(context.sourceAddress, context.client), int(context.sourceBalance))


@when('we mint NFT with invalid issuer')
def step_impl(context):
    try:
        payload = ObjFactory.getObj(
            ObjType.nft_token_mint,
            account=context.sourceAddress,
            issuer=context.invalidAddress,
            nftoken_taxon=0,
        )
        context.nft_mint_response_1 = sign_autofill_and_submit(context.client, payload, context.sourceWallet)
    except Exception as e:
        context.exception = e
        context.test_status = "failed"


@then('nft minting with invalid issuer is not successful')
def step_impl(context):
    verify_test_status(context.test_status, "failed")
    verify_test_error_message(str(context.exception), "expected valid classic address or X-Address, received str.")
    verify_account_balance(get_balance(context.sourceAddress, context.client), int(context.sourceBalance))


@when('we mint NFT with issuer having unauthorized user')
def step_impl(context):
    try:
        payload = AccountSet(
            account=context.sourceAddress,
            sequence=get_account_sequence(context.client, context.sourceAddress),
            fee="10",
            nftoken_minter=context.destinationAddress
        )
        context.account_set_response_1 = sign_autofill_and_submit(context.client, payload, context.sourceWallet)

        payload = ObjFactory.getObj(
            ObjType.nft_token_mint,
            account=context.destinationAddress,
            issuer=context.sourceAddress,
            nftoken_taxon=0,
        )
        context.nft_mint_response_1 = sign_autofill_and_submit(context.client, payload, context.destinationWallet)
    except Exception as e:
        context.exception = e
        context.test_status = "failed"


@then('nft minting with issuer having unauthorized user is not successful')
def step_impl(context):
    verify_test_status(context.test_status, "failed")
    verify_test_error_message(str(context.exception),
                              "'nftoken_minter': 'Will not set the minter unless AccountSetAsfFlag.ASF_AUTHORIZED_NFTOKEN_MINTER is set'")
    verify_account_balance(get_balance(context.sourceAddress, context.client), int(context.sourceBalance))


@when('we mint NFT with issuer having authorized user')
def step_impl(context):
    try:
        payload = AccountSet(
            account=context.sourceAddress,
            sequence=get_account_sequence(context.client, context.sourceAddress),
            fee="10",
            set_flag=AccountSetAsfFlag.ASF_AUTHORIZED_NFTOKEN_MINTER,
            nftoken_minter=context.destinationAddress
        )
        context.account_set_response_1 = sign_autofill_and_submit(context.client, payload, context.sourceWallet)

        payload = ObjFactory.getObj(
            ObjType.nft_token_mint,
            account=context.destinationAddress,
            issuer=context.sourceAddress,
            nftoken_taxon=0,
        )
        context.nft_mint_response_1 = sign_autofill_and_submit(context.client, payload, context.destinationWallet)
    except Exception as e:
        context.exception = e
        context.test_status = "failed"
        raise e


@then('nft minting with issuer having authorized user is successful')
def step_impl(context):
    verify_account_balance(get_balance(context.sourceAddress, context.client),
                           int(context.sourceBalance) - (context.minDrop))
    verify_account_balance(get_balance(context.destinationAddress, context.client),
                           int(context.destinationBalance) - (context.minDrop))
    verify_nftoken_mint_account_objects(context, context.nft_mint_response_1)

    verify_account_set_account_objects(context, context.account_set_response_1)


@when('we mint NFT on ticket with authorized user')
def step_impl(context):
    try:
        payload = AccountSet(
            account=context.sourceAddress,
            sequence=get_account_sequence(context.client, context.sourceAddress),
            fee="10",
            set_flag=AccountSetAsfFlag.ASF_AUTHORIZED_NFTOKEN_MINTER,
            nftoken_minter=context.destinationAddress
        )
        context.account_set_response_1 = sign_autofill_and_submit(context.client, payload, context.sourceWallet)

        payload = ObjFactory.getObj(
            ObjType.ticket_create,
            account=context.destinationAddress,
            sequence=get_account_sequence(context.client, context.destinationAddress),
            ticket_count=1,
        )
        context.tc_response_1 = sign_autofill_and_submit(context.client, payload, context.destinationWallet)

        ticket_seq = get_ticket_sequence(context.client, context.destinationAddress)[0]
        payload = ObjFactory.getObj(
            ObjType.nft_token_mint,
            account=context.destinationAddress,
            issuer=context.sourceAddress,
            nftoken_taxon=0,
            sequence=0,
            ticket_sequence=ticket_seq
        )

        context.nft_mint_response_1 = sign_autofill_and_submit(context.client, payload, context.destinationWallet)
    except Exception as e:
        context.exception = e
        context.test_status = "failed"
        raise e


@then('nft minting on ticket with authorized user is successful')
def step_impl(context):
    verify_account_balance(get_balance(context.sourceAddress, context.client),
                           int(context.sourceBalance) - (context.minDrop))
    verify_account_balance(get_balance(context.destinationAddress, context.client),
                           int(context.destinationBalance) - (2 * context.minDrop))

    verify_ticket_create_account_objects(context, context.tc_response_1)

    verify_nftoken_mint_account_objects(context, context.nft_mint_response_1)

    verify_account_set_account_objects(context, context.account_set_response_1)


@when('we mint NFT on ticket with issuer without user authorization')
def step_impl(context):
    try:
        payload = AccountSet(
            account=context.sourceAddress,
            set_flag=AccountSetAsfFlag.ASF_AUTHORIZED_NFTOKEN_MINTER,
            nftoken_minter=context.destinationAddress
        )
        context.account_set_response_1 = sign_autofill_and_submit(context.client, payload, context.sourceWallet)

        payload = AccountSet(
            account=context.sourceAddress,
            clear_flag=AccountSetAsfFlag.ASF_AUTHORIZED_NFTOKEN_MINTER,
        )
        context.account_set_response_2 = sign_autofill_and_submit(context.client, payload, context.sourceWallet)

        payload = ObjFactory.getObj(
            ObjType.nft_token_mint,
            account=context.destinationAddress,
            issuer=context.sourceAddress,
            nftoken_taxon=0,
        )
        context.nft_mint_response_1 = sign_autofill_and_submit(context.client, payload, context.destinationWallet)
    except Exception as e:
        context.exception = e
        context.test_status = "failed"


@then('nft minting on ticket with issuer without user authorization is not successful')
def step_impl(context):
    verify_test_status(context.test_status, "failed")
    verify_test_error_message(str(context.exception), "Transaction failed: tecNO_PERMISSION")
    verify_account_balance(get_balance(context.sourceAddress, context.client),
                           int(context.sourceBalance) - (2 * context.minDrop))
    verify_account_balance(get_balance(context.destinationAddress, context.client),
                           int(context.destinationBalance) - (context.minDrop))

    verify_account_set_account_objects(context, context.account_set_response_1)
    verify_account_set_account_objects(context, context.account_set_response_2)


@when('we mint NFT with issuer and remove authorization')
def step_impl(context):
    try:
        payload = AccountSet(
            account=context.sourceAddress,
            set_flag=AccountSetAsfFlag.ASF_AUTHORIZED_NFTOKEN_MINTER,
            nftoken_minter=context.destinationAddress
        )
        context.account_set_response_1 = sign_autofill_and_submit(context.client, payload, context.sourceWallet)

        payload = ObjFactory.getObj(
            ObjType.nft_token_mint,
            account=context.destinationAddress,
            issuer=context.sourceAddress,
            nftoken_taxon=0,
        )
        context.nft_mint_response_1 = sign_autofill_and_submit(context.client, payload, context.destinationWallet)

        payload = AccountSet(
            account=context.sourceAddress,
            clear_flag=AccountSetAsfFlag.ASF_AUTHORIZED_NFTOKEN_MINTER,
        )
        context.account_set_response_2 = sign_autofill_and_submit(context.client, payload, context.sourceWallet)

        payload = ObjFactory.getObj(
            ObjType.nft_token_mint,
            account=context.destinationAddress,
            issuer=context.sourceAddress,
            nftoken_taxon=0,
        )
        context.nft_mint_response_2 = sign_autofill_and_submit(context.client, payload, context.destinationWallet)
    except Exception as e:
        context.exception = e
        context.test_status = "failed"


@then('nft minting with issuer and removing authorization is successful')
def step_impl(context):
    assert_that(context.test_status, equal_to("failed"), "Test Status is incorrect")
    assert_that(str(context.exception), contains_string(
        "Transaction failed: tecNO_PERMISSION"))
    verify_account_balance(get_balance(context.sourceAddress, context.client),
                           int(context.sourceBalance) - (2 * context.minDrop))
    verify_account_balance(get_balance(context.destinationAddress, context.client),
                           int(context.destinationBalance) - (2 * context.minDrop))

    verify_nftoken_mint_account_objects(context, context.nft_mint_response_1)

    verify_account_set_account_objects(context, context.account_set_response_1)
    verify_account_set_account_objects(context, context.account_set_response_2)


@when('we change authorized user and mint NFT')
def step_impl(context):
    try:
        payload = AccountSet(
            account=context.aliceAddress,
            set_flag=AccountSetAsfFlag.ASF_AUTHORIZED_NFTOKEN_MINTER,
            nftoken_minter=context.bobAddress
        )
        context.account_set_response_1 = sign_autofill_and_submit(context.client, payload, context.aliceWallet)

        payload = AccountSet(
            account=context.aliceAddress,
            set_flag=AccountSetAsfFlag.ASF_AUTHORIZED_NFTOKEN_MINTER,
            nftoken_minter=context.carolAddress
        )
        context.account_set_response_2 = sign_autofill_and_submit(context.client, payload, context.aliceWallet)

        payload = ObjFactory.getObj(
            ObjType.nft_token_mint,
            account=context.bobAddress,
            issuer=context.aliceAddress,
            nftoken_taxon=0,
        )
        context.nft_mint_response_1 = sign_autofill_and_submit(context.client, payload, context.bobWallet)

        payload = ObjFactory.getObj(
            ObjType.nft_token_mint,
            account=context.carolAddress,
            issuer=context.aliceAddress,
            nftoken_taxon=0,
        )
        context.nft_mint_response_2 = sign_autofill_and_submit(context.client, payload, context.carolWallet)
    except Exception as e:
        context.exception = e
        context.test_status = "failed"


@then('NFT minting after changing authorized user is successful')
def step_impl(context):
    verify_test_status(context.test_status, "failed")
    verify_test_error_message(str(context.exception), "Transaction failed: tecNO_PERMISSION")

    verify_account_balance(get_balance(context.aliceAddress, context.client),
                           int(context.aliceBalance) - (2 * context.minDrop))
    verify_account_balance(get_balance(context.bobAddress, context.client),
                           int(context.bobBalance) - (context.minDrop))
    verify_account_balance(get_balance(context.carolAddress, context.client),
                           int(context.carolBalance))

    verify_account_set_account_objects(context, context.account_set_response_1)
    verify_account_set_account_objects(context, context.account_set_response_2)


@when('we mint NFT using authorization chain of authorized users')
def step_impl(context):
    try:
        payload = AccountSet(
            account=context.aliceAddress,
            set_flag=AccountSetAsfFlag.ASF_AUTHORIZED_NFTOKEN_MINTER,
            nftoken_minter=context.bobAddress
        )
        context.account_set_response_1 = sign_autofill_and_submit(context.client, payload, context.aliceWallet)

        payload = AccountSet(
            account=context.bobAddress,
            set_flag=AccountSetAsfFlag.ASF_AUTHORIZED_NFTOKEN_MINTER,
            nftoken_minter=context.carolAddress
        )
        context.account_set_response_2 = sign_autofill_and_submit(context.client, payload, context.bobWallet)

        payload = ObjFactory.getObj(
            ObjType.nft_token_mint,
            account=context.bobAddress,
            issuer=context.aliceAddress,
            nftoken_taxon=0,
        )
        context.nft_mint_response_1 = sign_autofill_and_submit(context.client, payload, context.bobWallet)

        payload = ObjFactory.getObj(
            ObjType.nft_token_mint,
            account=context.carolAddress,
            issuer=context.aliceAddress,
            nftoken_taxon=0,
        )
        context.nft_mint_response_2 = sign_autofill_and_submit(context.client, payload, context.carolWallet)
    except Exception as e:
        context.exception = e
        context.test_status = "failed"


@then('NFT minting using authorization chain of authorized users is not successful')
def step_impl(context):
    verify_test_status(context.test_status, "failed")
    verify_test_error_message(str(context.exception), "Transaction failed: tecNO_PERMISSION")

    verify_account_balance(get_balance(context.aliceAddress, context.client),
                           int(context.aliceBalance) - (context.minDrop))
    verify_account_balance(get_balance(context.bobAddress, context.client),
                           int(context.bobBalance) - (2 * context.minDrop))
    verify_account_balance(get_balance(context.carolAddress, context.client),
                           int(context.carolBalance) - (context.minDrop))

    verify_nftoken_mint_account_objects(context, context.nft_mint_response_1)

    verify_account_set_account_objects(context, context.account_set_response_1)
    verify_account_set_account_objects(context, context.account_set_response_2)


@when('we mint NFT using ticket')
def step_impl(context):
    try:
        payload = ObjFactory.getObj(
            ObjType.ticket_create,
            account=context.sourceAddress,
            sequence=get_account_sequence(context.client, context.sourceAddress),
            ticket_count=1,
        )
        context.tc_response_1 = sign_autofill_and_submit(context.client, payload, context.sourceWallet)

        payload = ObjFactory.getObj(
            ObjType.nft_token_mint,
            account=context.sourceAddress,
            nftoken_taxon=0,
            sequence=0,
            ticket_sequence=get_ticket_sequence(context.client, context.sourceAddress)[0]
        )
        context.nft_mint_response_1 = sign_autofill_and_submit(context.client, payload, context.sourceWallet)
    except Exception as e:
        context.exception = e
        context.test_status = "failed"
        raise e


@then('NFT minting using ticket is successful')
def step_impl(context):
    verify_account_balance(get_balance(context.sourceAddress, context.client),
                           int(context.sourceBalance) - (2 * context.minDrop))

    verify_ticket_create_account_objects(context, context.tc_response_1)
    verify_nftoken_mint_account_objects(context, context.nft_mint_response_1)


@when('we mint NFT and try to delete account owner')
def step_impl(context):
    try:
        payload = ObjFactory.getObj(
            ObjType.nft_token_mint,
            account=context.sourceAddress,
            nftoken_taxon=0,
        )
        context.nft_mint_response_1 = sign_autofill_and_submit(context.client, payload, context.sourceWallet)

        payload = ObjFactory.getObj(
            ObjType.account_delete,
            account=context.sourceAddress,
            destination=context.destinationAddress,
            fee=constants.DEFAULT_DELETE_ACCOUNT_FEE,
            sequence=get_account_sequence(context.client, context.sourceAddress)
        )
        response = sign_autofill_and_submit(context.client, payload, context.sourceWallet)
        context.ad_response_1 = context.client.request(Tx(transaction=response.result["hash"]))
    except Exception as e:
        context.exception = e
        context.test_status = "failed"


@then('deleting account owner after NFT minting is not successful')
def step_impl(context):
    verify_test_status(context.test_status, "failed")
    verify_test_error_message(str(context.exception), "Transaction failed: tecHAS_OBLIGATIONS")
    verify_account_balance(get_balance(context.sourceAddress, context.client),
                           int(context.sourceBalance) - (context.minDrop) - int(constants.DEFAULT_DELETE_ACCOUNT_FEE))
    verify_account_balance(get_balance(context.destinationAddress, context.client),
                           int(context.destinationBalance))
    verify_nftoken_mint_account_objects(context, context.nft_mint_response_1)


@when('we mint NFT on ticket and try to delete account owner')
def step_impl(context):
    try:
        payload = ObjFactory.getObj(
            ObjType.ticket_create,
            account=context.sourceAddress,
            sequence=get_account_sequence(context.client, context.sourceAddress),
            ticket_count=1,
        )
        context.tc_response_1 = sign_autofill_and_submit(context.client, payload, context.sourceWallet)

        payload = ObjFactory.getObj(
            ObjType.nft_token_mint,
            account=context.sourceAddress,
            nftoken_taxon=0,
            sequence=0,
            ticket_sequence=get_ticket_sequence(context.client, context.sourceAddress)[0]
        )
        context.nft_mint_response_1 = sign_autofill_and_submit(context.client, payload, context.sourceWallet)

        payload = ObjFactory.getObj(
            ObjType.account_delete,
            account=context.sourceAddress,
            destination=context.destinationAddress,
            fee=constants.DEFAULT_DELETE_ACCOUNT_FEE,
            sequence=get_account_sequence(context.client, context.sourceAddress)
        )
        response = sign_autofill_and_submit(context.client, payload, context.sourceWallet)
        context.ad_response_1 = context.client.request(Tx(transaction=response.result["hash"]))
    except Exception as e:
        context.exception = e
        context.test_status = "failed"


@then('deleting account owner after NFT minting on ticket is not successful')
def step_impl(context):
    verify_test_status(context.test_status, "failed")
    verify_test_error_message(str(context.exception), "Transaction failed: tecHAS_OBLIGATIONS")

    verify_account_balance(get_balance(context.sourceAddress, context.client),
                           int(context.sourceBalance) - (2 * context.minDrop) - int(constants.DEFAULT_DELETE_ACCOUNT_FEE))
    verify_account_balance(get_balance(context.destinationAddress, context.client),
                           int(context.destinationBalance))

    verify_ticket_create_account_objects(context, context.tc_response_1)
    verify_nftoken_mint_account_objects(context, context.nft_mint_response_1)


@when('we mint NFT with issuer and then remove authorization and delete account owner')
def step_impl(context):
    try:

        payload = AccountSet(
            account=context.sourceAddress,
            set_flag=AccountSetAsfFlag.ASF_AUTHORIZED_NFTOKEN_MINTER,
            nftoken_minter=context.destinationAddress
        )
        context.account_set_response_1 = sign_autofill_and_submit(context.client, payload, context.sourceWallet)

        payload = ObjFactory.getObj(
            ObjType.nft_token_mint,
            account=context.destinationAddress,
            issuer=context.sourceAddress,
            nftoken_taxon=0,
        )
        context.nft_mint_response_1 = sign_autofill_and_submit(context.client, payload, context.destinationWallet)

        payload = AccountSet(
            account=context.sourceAddress,
            clear_flag=AccountSetAsfFlag.ASF_AUTHORIZED_NFTOKEN_MINTER,
        )
        context.account_set_response_2 = sign_autofill_and_submit(context.client, payload, context.sourceWallet)

        payload = ObjFactory.getObj(
            ObjType.account_delete,
            account=context.destinationAddress,
            destination=context.sourceAddress,
            fee=constants.DEFAULT_DELETE_ACCOUNT_FEE,
            sequence=get_account_sequence(context.client, context.destinationAddress)
        )
        context.ad_response_1 = sign_autofill_and_submit(context.client, payload, context.destinationWallet)
    except Exception as e:
        context.exception = e
        context.test_status = "failed"


@then('removing authorization and deleting account owner after NFT minting is not successful')
def step_impl(context):
    verify_test_status(context.test_status, "failed")
    verify_test_error_message(str(context.exception), "Transaction failed: tecHAS_OBLIGATIONS")
    verify_account_balance(get_balance(context.sourceAddress, context.client),
                           int(context.sourceBalance) - (2 * context.minDrop))
    verify_account_balance(get_balance(context.destinationAddress, context.client),
                           int(context.destinationBalance) - context.minDrop - int(constants.DEFAULT_DELETE_ACCOUNT_FEE))
    verify_nftoken_mint_account_objects(context, context.nft_mint_response_1)
    verify_account_set_account_objects(context, context.account_set_response_1)
    verify_account_set_account_objects(context, context.account_set_response_2)


@when('we mint NFT with issuer and then remove authorization and delete issuer')
def step_impl(context):
    try:
        payload = AccountSet(
            account=context.sourceAddress,
            set_flag=AccountSetAsfFlag.ASF_AUTHORIZED_NFTOKEN_MINTER,
            nftoken_minter=context.destinationAddress
        )
        context.account_set_response_1 = sign_autofill_and_submit(context.client, payload, context.sourceWallet)

        payload = ObjFactory.getObj(
            ObjType.nft_token_mint,
            account=context.destinationAddress,
            issuer=context.sourceAddress,
            nftoken_taxon=0,
        )
        context.nft_mint_response_1 = sign_autofill_and_submit(context.client, payload, context.destinationWallet)

        payload = AccountSet(
            account=context.sourceAddress,
            clear_flag=AccountSetAsfFlag.ASF_AUTHORIZED_NFTOKEN_MINTER,
        )
        context.account_set_response_2 = sign_autofill_and_submit(context.client, payload, context.sourceWallet)

        payload = ObjFactory.getObj(
            ObjType.account_delete,
            account=context.sourceAddress,
            destination=context.destinationAddress,
            fee=constants.DEFAULT_DELETE_ACCOUNT_FEE,
            sequence=get_account_sequence(context.client, context.sourceAddress)
        )
        context.ad_response_1 = sign_autofill_and_submit(context.client, payload, context.sourceWallet)
    except Exception as e:
        context.exception = e
        context.test_status = "failed"


@then('removing authorization and deleting issuer after NFT minting is not successful')
def step_impl(context):
    verify_test_status(context.test_status, "failed")
    verify_test_error_message(str(context.exception), "Transaction failed: tecHAS_OBLIGATIONS")
    verify_account_balance(get_balance(context.sourceAddress, context.client),
                           int(context.sourceBalance) - (2 * context.minDrop) - int(constants.DEFAULT_DELETE_ACCOUNT_FEE))
    verify_account_balance(get_balance(context.destinationAddress, context.client),
                           int(context.destinationBalance) - context.minDrop)

    verify_nftoken_mint_account_objects(context, context.nft_mint_response_1)
    verify_account_set_account_objects(context, context.account_set_response_1)
    verify_account_set_account_objects(context, context.account_set_response_2)


@when('we burn NFT as owner')
def step_impl(context):
    try:
        payload = ObjFactory.getObj(
            ObjType.nft_token_mint,
            account=context.sourceAddress,
            nftoken_taxon=0,
        )
        context.nft_mint_response_1 = sign_autofill_and_submit(context.client, payload, context.sourceWallet)

        payload = ObjFactory.getObj(
            ObjType.nft_token_burn,
            account=context.sourceAddress,
            nftoken_id=get_nft_tokens(context.client, context.sourceAddress)[-1],
        )
        context.nft_burn_response_1 = sign_autofill_and_submit(context.client, payload, context.sourceWallet)
    except Exception as e:
        context.exception = e
        context.test_status = "failed"
        raise e


@then('burning NFT as owner is successful')
def step_impl(context):
    verify_account_balance(get_balance(context.sourceAddress, context.client),
                           int(context.sourceBalance) - (2 * context.minDrop))
    verify_nftoken_mint_account_objects(context, context.nft_mint_response_1)
    verify_nftoken_burn_account_objects(context, context.nft_burn_response_1)


@when('we burn NFT with low reserves')
def step_impl(context):
    try:
        payload = ObjFactory.getObj(
            ObjType.nft_token_mint,
            account=context.sourceAddress,
            nftoken_taxon=0,
        )
        context.nft_mint_response_1 = sign_autofill_and_submit(context.client, payload, context.sourceWallet)

        final_transfer_amount = str(
            get_balance(context.sourceAddress, context.client) - int(constants.BASE_RESERVE) - int(
                constants.OWNER_RESERVE) - context.minDrop)
        context.py_response_1 = sign_autofill_and_submit(context.client,
                                                         ObjFactory.getObj(
                                                             ObjType.payment,
                                                             account=context.sourceAddress,
                                                             destination=context.destinationAddress,
                                                             amount=final_transfer_amount),
                                                         context.sourceWallet)

        payload = ObjFactory.getObj(
            ObjType.nft_token_burn,
            account=context.sourceAddress,
            nftoken_id=get_nft_tokens(context.client, context.sourceAddress)[-1],
        )
        context.nft_burn_response_1 = sign_autofill_and_submit(context.client, payload, context.sourceWallet)
    except Exception as e:
        context.exception = e
        context.test_status = "failed"
        raise e


@then('burning NFT with low reserves is not successful')
def step_impl(context):
    verify_nftoken_mint_account_objects(context, context.nft_mint_response_1)
    verify_nftoken_burn_account_objects(context, context.nft_burn_response_1)
    verify_payment_account_objects(context, context.py_response_1)


@when('we burn NFT with NFT ID mismatch')
def step_impl(context):
    try:
        payload = ObjFactory.getObj(
            ObjType.nft_token_burn,
            account=context.sourceAddress,
            fee="10",
            nftoken_id="0008000044CAF362635003E9D565979EE87A1668A1FFE7BB2DCBAB9D00000002",
        )
        context.nft_burn_response_1 = sign_autofill_and_submit(context.client, payload, context.sourceWallet)
    except Exception as e:
        context.exception = e
        context.test_status = "failed"


@then('burning NFT with NFT ID mismatch is not successful')
def step_impl(context):
    verify_test_status(context.test_status, "failed")
    verify_test_error_message(str(context.exception), "Transaction failed: tecNO_ENTRY")

    verify_account_balance(get_balance(context.sourceAddress, context.client),
                           int(context.sourceBalance) - context.minDrop)


@when('we burn NFT as different user')
def step_impl(context):
    try:
        payload = ObjFactory.getObj(
            ObjType.nft_token_mint,
            account=context.sourceAddress,
            nftoken_taxon=0,
        )
        context.nft_mint_response_1 = sign_autofill_and_submit(context.client, payload, context.sourceWallet)

        payload = ObjFactory.getObj(
            ObjType.nft_token_burn,
            account=context.destinationAddress,
            fee="10",
            nftoken_id=get_nft_tokens(context.client, context.sourceAddress)[-1],
        )
        context.nft_burn_response_1 = sign_autofill_and_submit(context.client, payload, context.destinationWallet)
    except Exception as e:
        context.exception = e
        context.test_status = "failed"


@then('burning NFT as different user is not successful')
def step_impl(context):
    verify_test_status(context.test_status, "failed")
    verify_test_error_message(str(context.exception), "Transaction failed: tecNO_ENTRY")

    verify_account_balance(get_balance(context.sourceAddress, context.client),
                           int(context.sourceBalance) - context.minDrop)
    verify_account_balance(get_balance(context.destinationAddress, context.client),
                           int(context.destinationBalance) - context.minDrop)
    verify_nftoken_mint_account_objects(context, context.nft_mint_response_1)


@when('we mint NFT with issuer burn as owner')
def step_impl(context):
    try:
        payload = AccountSet(
            account=context.sourceAddress,
            set_flag=AccountSetAsfFlag.ASF_AUTHORIZED_NFTOKEN_MINTER,
            nftoken_minter=context.destinationAddress
        )
        context.account_set_response_1 = sign_autofill_and_submit(context.client, payload, context.sourceWallet)
        #
        payload = ObjFactory.getObj(
            ObjType.nft_token_mint,
            account=context.destinationAddress,
            issuer=context.sourceAddress,
            nftoken_taxon=0,
        )
        context.nft_mint_response_1 = sign_autofill_and_submit(context.client, payload, context.destinationWallet)

        payload = ObjFactory.getObj(
            ObjType.nft_token_burn,
            account=context.destinationAddress,
            nftoken_id=get_nft_tokens(context.client, context.destinationAddress)[-1],
        )
        context.nft_burn_response_1 = sign_autofill_and_submit(context.client, payload, context.destinationWallet)
    except Exception as e:
        context.exception = e
        context.test_status = "failed"
        raise e


@then('burning NFT with issuer burn as owner is successful')
def step_impl(context):
    verify_account_balance(get_balance(context.sourceAddress, context.client),
                           int(context.sourceBalance) - (context.minDrop))

    verify_account_balance(get_balance(context.destinationAddress, context.client),
                           int(context.destinationBalance) - (2 * context.minDrop))

    verify_nftoken_mint_account_objects(context, context.nft_mint_response_1)
    verify_nftoken_burn_account_objects(context, context.nft_burn_response_1)

    verify_account_set_account_objects(context, context.account_set_response_1)


@when('we mint NFT with issuer burn as issuer without tfBurnable')
def step_impl(context):
    try:
        payload = AccountSet(
            account=context.sourceAddress,
            set_flag=AccountSetAsfFlag.ASF_AUTHORIZED_NFTOKEN_MINTER,
            nftoken_minter=context.destinationAddress
        )
        context.account_set_response_1 = sign_autofill_and_submit(context.client, payload, context.sourceWallet)
        #
        payload = ObjFactory.getObj(
            ObjType.nft_token_mint,
            issuer=context.sourceAddress,
            account=context.destinationAddress,
            nftoken_taxon=0,
        )
        context.nft_mint_response_1 = sign_autofill_and_submit(context.client, payload, context.destinationWallet)

        payload = ObjFactory.getObj(
            ObjType.nft_token_burn,
            account=context.sourceAddress,
            owner=context.destinationAddress,
            fee="10",
            nftoken_id=get_nft_tokens(context.client, context.destinationAddress)[-1],
        )
        context.nft_burn_response_1 = sign_autofill_and_submit(context.client, payload, context.sourceWallet)
    except Exception as e:
        context.exception = e
        context.test_status = "failed"


@then('burning NFT with issuer burn as issuer without tfBurnable is not successful')
def step_impl(context):
    verify_test_status(context.test_status, "failed")
    verify_test_error_message(str(context.exception), "Transaction failed: tecNO_PERMISSION")

    verify_account_balance(get_balance(context.sourceAddress, context.client),
                           int(context.sourceBalance) - (2 * context.minDrop))
    verify_account_balance(get_balance(context.destinationAddress, context.client),
                           int(context.destinationBalance) - context.minDrop)
    verify_nftoken_mint_account_objects(context, context.nft_mint_response_1)
    verify_account_set_account_objects(context, context.account_set_response_1)


@when('we mint NFT with issuer burn as issuer with tfBurnable')
def step_impl(context):
    try:
        payload = AccountSet(
            account=context.sourceAddress,
            set_flag=AccountSetAsfFlag.ASF_AUTHORIZED_NFTOKEN_MINTER,
            nftoken_minter=context.destinationAddress
        )
        context.account_set_response_1 = sign_autofill_and_submit(context.client, payload, context.sourceWallet)
        #
        payload = ObjFactory.getObj(
            ObjType.nft_token_mint,
            account=context.destinationAddress,
            issuer=context.sourceAddress,
            nftoken_taxon=0,
            flags=1
        )
        context.nft_mint_response_1 = sign_autofill_and_submit(context.client, payload, context.destinationWallet)

        payload = ObjFactory.getObj(
            ObjType.nft_token_burn,
            account=context.sourceAddress,
            owner=context.destinationAddress,
            fee="10",
            nftoken_id=get_nft_tokens(context.client, context.destinationAddress)[-1],
        )
        context.nft_burn_response_1 = sign_autofill_and_submit(context.client, payload, context.sourceWallet)
    except Exception as e:
        context.exception = e
        context.test_status = "failed"
        raise e


@then('burning NFT with issuer burn as issuer with tfBurnable is successful')
def step_impl(context):
    verify_account_balance(get_balance(context.sourceAddress, context.client),
                           int(context.sourceBalance) - (2 * context.minDrop))
    verify_account_balance(get_balance(context.destinationAddress, context.client),
                           int(context.destinationBalance) - context.minDrop)
    verify_nftoken_mint_account_objects(context, context.nft_mint_response_1)
    verify_nftoken_burn_account_objects(context, context.nft_burn_response_1)
    verify_account_set_account_objects(context, context.account_set_response_1)


@when('we mint NFT with issuer burn as issuer without owner field')
def step_impl(context):
    try:
        payload = AccountSet(
            account=context.sourceAddress,
            set_flag=AccountSetAsfFlag.ASF_AUTHORIZED_NFTOKEN_MINTER,
            nftoken_minter=context.destinationAddress
        )
        context.account_set_response_1 = sign_autofill_and_submit(context.client, payload, context.sourceWallet)
        #
        payload = ObjFactory.getObj(
            ObjType.nft_token_mint,
            issuer=context.sourceAddress,
            account=context.destinationAddress,
            nftoken_taxon=0,
            flags=1
        )
        context.nft_mint_response_1 = sign_autofill_and_submit(context.client, payload, context.destinationWallet)

        payload = ObjFactory.getObj(
            ObjType.nft_token_burn,
            account=context.sourceAddress,
            fee="10",
            nftoken_id=get_nft_tokens(context.client, context.destinationAddress)[-1],
        )
        context.nft_burn_response_1 = sign_autofill_and_submit(context.client, payload, context.sourceWallet)
    except Exception as e:
        context.exception = e
        context.test_status = "failed"


@then('burning NFT with issuer burn as issuer without owner field not successful')
def step_impl(context):
    verify_test_status(context.test_status, "failed")
    verify_test_error_message(str(context.exception), "Transaction failed: tecNO_ENTRY")
    verify_account_balance(get_balance(context.sourceAddress, context.client),
                           int(context.sourceBalance) - (2 * context.minDrop))
    verify_account_balance(get_balance(context.destinationAddress, context.client),
                           int(context.destinationBalance) - context.minDrop)
    verify_nftoken_mint_account_objects(context, context.nft_mint_response_1)
    verify_account_set_account_objects(context, context.account_set_response_1)


@when('we burn NFT and remint')
def step_impl(context):
    try:
        payload = ObjFactory.getObj(
            ObjType.nft_token_mint,
            account=context.sourceAddress,
            nftoken_taxon=0,
        )
        context.nft_mint_response_1 = sign_autofill_and_submit(context.client, payload, context.sourceWallet)

        payload = ObjFactory.getObj(
            ObjType.nft_token_burn,
            account=context.sourceAddress,
            nftoken_id=get_nft_tokens(context.client, context.sourceAddress)[-1],
        )
        context.nft_burn_response_1 = sign_autofill_and_submit(context.client, payload, context.sourceWallet)

        payload = ObjFactory.getObj(
            ObjType.nft_token_mint,
            account=context.sourceAddress,
            nftoken_taxon=0,
        )
        context.nft_mint_response_2 = sign_autofill_and_submit(context.client, payload, context.sourceWallet)
    except Exception as e:
        context.exception = e
        context.test_status = "failed"
        raise e


@then('burning NFT and reminting is successful')
def step_impl(context):
    verify_account_balance(get_balance(context.sourceAddress, context.client),
                           int(context.sourceBalance) - (3 * context.minDrop))

    verify_nftoken_mint_account_objects(context, context.nft_mint_response_1)
    verify_nftoken_mint_account_objects(context, context.nft_mint_response_2)
    verify_nftoken_burn_account_objects(context, context.nft_burn_response_1)


@when('we burn NFT and delete owner')
def step_impl(context):
    try:
        payload = ObjFactory.getObj(
            ObjType.nft_token_mint,
            account=context.sourceAddress,
            nftoken_taxon=0,
        )
        context.nft_mint_response_1 = sign_autofill_and_submit(context.client, payload, context.sourceWallet)

        payload = ObjFactory.getObj(
            ObjType.nft_token_burn,
            account=context.sourceAddress,
            nftoken_id=get_nft_tokens(context.client, context.sourceAddress)[-1],
        )
        context.nft_burn_response_1 = sign_autofill_and_submit(context.client, payload, context.sourceWallet)

        wait_for_ledger_to_advance_for_account_delete(context, context.sourceAddress)

        payload = ObjFactory.getObj(
            ObjType.account_delete,
            account=context.sourceAddress,
            destination=context.destinationAddress,
            fee=constants.DEFAULT_DELETE_ACCOUNT_FEE,
            sequence=get_account_sequence(context.client, context.sourceAddress)
        )
        context.ad_response_1 = sign_autofill_and_submit(context.client, payload, context.sourceWallet)

    except Exception as e:
        context.exception = e
        context.test_status = "failed"

        raise e


@then('burning NFT and deleting owner is successful')
def step_impl(context):
    verify_nftoken_mint_account_objects(context, context.nft_mint_response_1)
    verify_account_delete_account_objects(context, context.ad_response_1)
    verify_nftoken_burn_account_objects(context, context.nft_burn_response_1)


@when('we mint NFT with issuer burn as owner and delete owner')
def step_impl(context):
    try:
        payload = AccountSet(
            account=context.sourceAddress,
            set_flag=AccountSetAsfFlag.ASF_AUTHORIZED_NFTOKEN_MINTER,
            nftoken_minter=context.destinationAddress
        )
        context.account_set_response_1 = sign_autofill_and_submit(context.client, payload, context.sourceWallet)

        payload = ObjFactory.getObj(
            ObjType.nft_token_mint,
            account=context.destinationAddress,
            issuer=context.sourceAddress,
            nftoken_taxon=0,
        )
        context.nft_mint_response_1 = sign_autofill_and_submit(context.client, payload, context.destinationWallet)

        payload = ObjFactory.getObj(
            ObjType.nft_token_burn,
            account=context.destinationAddress,
            nftoken_id=get_nft_tokens(context.client, context.destinationAddress)[-1],
        )
        context.nft_burn_response_1 = sign_autofill_and_submit(context.client, payload, context.destinationWallet)

        wait_for_ledger_to_advance_for_account_delete(context, context.destinationAddress)

        payload = ObjFactory.getObj(
            ObjType.account_delete,
            account=context.destinationAddress,
            destination=context.sourceAddress,
            fee=constants.DEFAULT_DELETE_ACCOUNT_FEE,
            sequence=get_account_sequence(context.client, context.destinationAddress)
        )
        context.ad_response_1 = sign_autofill_and_submit(context.client, payload, context.destinationWallet)

    except Exception as e:
        context.exception = e
        context.test_status = "failed"
        raise e


@then('minting NFT with issuer burn as owner and deleting owner is successful')
def step_impl(context):
    verify_nftoken_mint_account_objects(context, context.nft_mint_response_1)
    verify_account_delete_account_objects(context, context.ad_response_1)
    verify_nftoken_burn_account_objects(context, context.nft_burn_response_1)
    verify_account_set_account_objects(context, context.account_set_response_1)


@when('we burn NFT as owner and delete issuer')
def step_impl(context):
    try:
        payload = AccountSet(
            account=context.sourceAddress,
            set_flag=AccountSetAsfFlag.ASF_AUTHORIZED_NFTOKEN_MINTER,
            nftoken_minter=context.destinationAddress
        )
        context.account_set_response_1 = sign_autofill_and_submit(context.client, payload, context.sourceWallet)

        payload = ObjFactory.getObj(
            ObjType.nft_token_mint,
            account=context.destinationAddress,
            issuer=context.sourceAddress,
            nftoken_taxon=0,
        )
        context.nft_mint_response_1 = sign_autofill_and_submit(context.client, payload, context.destinationWallet)

        payload = ObjFactory.getObj(
            ObjType.nft_token_burn,
            account=context.destinationAddress,
            nftoken_id=get_nft_tokens(context.client, context.destinationAddress)[-1],
        )
        context.nft_burn_response_1 = sign_autofill_and_submit(context.client, payload, context.destinationWallet)

        wait_for_ledger_to_advance_for_account_delete(context, context.sourceAddress)

        payload = ObjFactory.getObj(
            ObjType.account_delete,
            account=context.sourceAddress,
            destination=context.destinationAddress,
            fee=constants.DEFAULT_DELETE_ACCOUNT_FEE,
            sequence=get_account_sequence(context.client, context.sourceAddress)
        )
        context.ad_response_1 = sign_autofill_and_submit(context.client, payload, context.sourceWallet)

    except Exception as e:
        context.exception = e
        context.test_status = "failed"
        raise e


@then('burning NFT as owner and deleting issuer is successful')
def step_impl(context):
    verify_nftoken_mint_account_objects(context, context.nft_mint_response_1)
    verify_account_delete_account_objects(context, context.ad_response_1)
    verify_nftoken_burn_account_objects(context, context.nft_burn_response_1)
    verify_account_set_account_objects(context, context.account_set_response_1)


@when('we burn NFT as issuer and delete owner')
def step_impl(context):
    try:
        payload = AccountSet(
            account=context.sourceAddress,
            set_flag=AccountSetAsfFlag.ASF_AUTHORIZED_NFTOKEN_MINTER,
            nftoken_minter=context.destinationAddress
        )
        context.account_set_response_1 = sign_autofill_and_submit(context.client, payload, context.sourceWallet)
        #
        payload = ObjFactory.getObj(
            ObjType.nft_token_mint,
            account=context.destinationAddress,
            issuer=context.sourceAddress,
            nftoken_taxon=0,
            flags=1
        )
        context.nft_mint_response_1 = sign_autofill_and_submit(context.client, payload, context.destinationWallet)

        payload = ObjFactory.getObj(
            ObjType.nft_token_burn,
            account=context.sourceAddress,
            owner=context.destinationAddress,
            nftoken_id=get_nft_tokens(context.client, context.destinationAddress)[-1],
        )
        context.nft_burn_response_1 = sign_autofill_and_submit(context.client, payload, context.sourceWallet)

        wait_for_ledger_to_advance_for_account_delete(context, context.sourceAddress)

        payload = ObjFactory.getObj(
            ObjType.account_delete,
            account=context.destinationAddress,
            destination=context.sourceAddress,
            fee=constants.DEFAULT_DELETE_ACCOUNT_FEE,
            sequence=get_account_sequence(context.client, context.destinationAddress)
        )
        context.ad_response_1 = sign_autofill_and_submit(context.client, payload, context.destinationWallet)

    except Exception as e:
        context.exception = e
        context.test_status = "failed"

        raise e


@then('burning NFT as issuer and deleting owner is successful')
def step_impl(context):
    verify_nftoken_mint_account_objects(context, context.nft_mint_response_1)
    verify_account_delete_account_objects(context, context.ad_response_1)
    verify_nftoken_burn_account_objects(context, context.nft_burn_response_1)
    verify_account_set_account_objects(context, context.account_set_response_1)


@when('we burn NFT as issuer and delete issuer')
def step_impl(context):
    try:
        payload = AccountSet(
            account=context.sourceAddress,
            set_flag=AccountSetAsfFlag.ASF_AUTHORIZED_NFTOKEN_MINTER,
            nftoken_minter=context.destinationAddress
        )
        context.account_set_response_1 = sign_autofill_and_submit(context.client, payload, context.sourceWallet)

        payload = ObjFactory.getObj(
            ObjType.nft_token_mint,
            account=context.destinationAddress,
            issuer=context.sourceAddress,
            nftoken_taxon=0,
            flags=1
        )
        context.nft_mint_response_1 = sign_autofill_and_submit(context.client, payload, context.destinationWallet)

        payload = ObjFactory.getObj(
            ObjType.nft_token_burn,
            account=context.sourceAddress,
            owner=context.destinationAddress,
            nftoken_id=get_nft_tokens(context.client, context.destinationAddress)[-1],
        )
        context.nft_burn_response_1 = sign_autofill_and_submit(context.client, payload, context.sourceWallet)

        wait_for_ledger_to_advance_for_account_delete(context, context.sourceAddress)

        payload = ObjFactory.getObj(
            ObjType.account_delete,
            account=context.sourceAddress,
            destination=context.destinationAddress,
            fee=constants.DEFAULT_DELETE_ACCOUNT_FEE,
            sequence=get_account_sequence(context.client, context.sourceAddress)
        )
        context.ad_response_1 = sign_autofill_and_submit(context.client, payload, context.sourceWallet)

    except Exception as e:
        context.exception = e
        context.test_status = "failed"

        raise e


@then('burning NFT as issuer and deleting issuer is successful')
def step_impl(context):
    verify_nftoken_mint_account_objects(context, context.nft_mint_response_1)
    verify_account_delete_account_objects(context, context.ad_response_1)
    verify_nftoken_burn_account_objects(context, context.nft_burn_response_1)
    verify_account_set_account_objects(context, context.account_set_response_1)


@when('we mint more than 32 NFT objects')
def step_impl(context):
    try:
        payload = ObjFactory.getObj(
            ObjType.nft_token_mint,
            account=context.sourceAddress,
            nftoken_taxon=0,
        )
        max_nftokens = 35
        context.responses = []

        for count in range(1, max_nftokens + 1):
            context.responses.append(sign_autofill_and_submit(context.client, payload, context.sourceWallet))
    except Exception as e:
        context.exception = e
        context.test_status = "failed"
        raise e


@then('minting more than 32 NFT objects is successful')
def step_impl(context):
    for response in context.responses:
        verify_nftoken_mint_account_objects(context, response)



import os
import shlex
import subprocess

import streamlit as st
from PIL import Image
from claudia.aws.aws_helper import are_keys_valid
from claudia.claudia import *
from claudia.ui.ui_helper import create_account, send_payment_across_accounts, mint_nft, burn_nft, \
    create_nft_buy_offer, accept_nft_buy_offer, accept_nft_sell_offer, create_nft_sell_offer
from streamlit_option_menu import option_menu

from claudia.versions import XRPL_PY_VERSION, XRPL_JS_VERSION
from claudia.aws.aws_helper import create_ec2_instance

st.set_page_config(layout='wide')
set_to_project_root_wd()
st.title('Claudia')
st.caption('XRPL No-Code Automation Framework')

def execute(cmd):
    split_cmd = shlex.split(cmd)
    popen = subprocess.Popen(split_cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, universal_newlines=True)
    for stdout_line in iter(popen.stdout.readline, ""):
        yield stdout_line
    popen.stdout.close()
    return_code = popen.wait()
    if return_code:
        raise subprocess.CalledProcessError(return_code, split_cmd)


def launch_console_and_execute(cmd):
    st.markdown(
        '''
        <style>
        .streamlit-expanderHeader {
            background-color: #000000;
            color: #FFFFFF;
            font-family: "Lucida Console", "Courier New", monospace;
        }
        .streamlit-expanderContent {
            background-color: #000000;
            color: #FFFFFF;
            font-family: "Lucida Console", "Courier New", monospace;
        }
        </style>
        ''',
        unsafe_allow_html=True
    )

    with st.expander("Details", expanded=True):
        try:
            for path in execute(cmd):
                st.text(path)

            st.success('Successfully completed the task.')
        except Exception as e:
            st.error(f'Something went wrong.\nDetails: {str(e)}')


def launch_console_and_display(message):
    st.markdown(
        '''
        <style>
        .streamlit-expanderHeader {
            background-color: #000000;
            color: #FFFFFF;
            font-family: "Lucida Console", "Courier New", monospace;
        }
        .streamlit-expanderContent {
            background-color: #000000;
            color: #FFFFFF;
            font-family: "Lucida Console", "Courier New", monospace;
        }
        </style>
        ''',
        unsafe_allow_html=True
    )

    with st.expander("Details", expanded=True):
        try:
            st.text(message)
            st.success('Successfully completed the task.')
        except Exception as e:
            st.error(f'Something went wrong.\nDetails: {str(e)}')


def click_learning_path_network_button():
    st.toast("click_learning_path_network_button has been called")
    if 'clicked_learning_path_network_button' not in st.session_state:
        st.session_state.clicked_learning_path_network_button = True
    st.toast(
        f"st.session_state.clicked_learning_path_network_button: {st.session_state.clicked_learning_path_network_button}")


with st.sidebar:
    image = Image.open('./ui/logo.png')
    st.image(image)

    sidebar_selection = option_menu(
        "Claudia",
        ["Home", "Custom XRPL Networks", 'XRPL Tests', 'XRPL Learning Center', 'Settings'],
        icons=['house', 'cpu', "cpu", 'binoculars', 'book', 'gear'],
        menu_icon="balloon",
        default_index=0,
        key=1
    )

if sidebar_selection == "Home":
    with st.container():
        st.write("Welcome to Claudia!")
        st.markdown(
            'Claudia is a helper utility that allows users to perform XRPL specific tasks both locally and on other '
            'public networks. More information can be found at [Claudia\'s PyPi homepage]'
            '(https://pypi.org/project/claudia/). '
        )
        st.markdown(
            'This version uses:'
        )
        st.markdown(
            f' - **XRPL-py v{XRPL_PY_VERSION}**'
        )
        st.markdown(
            f' - **XRPL.js v{XRPL_JS_VERSION}**'
        )
        st.balloons()
elif sidebar_selection == "Custom XRPL Networks":
    with st.container():
        network_menu = option_menu(
            sidebar_selection,
            [
                "Build rippled",
                "Install rippled",
                "Start Network",
                "Stop Network",
                "Network Status",
                "Build Witness Server",
                "Start Sidechain Network",
                "Stop Sidechain Network",
                "Check Witness Server Status",
                "Deploy Network to Cloud"
            ],
            icons=[
                'building-gear',
                'building-check',
                'play',
                'stop',
                'eyeglasses',
                'building-gear',
                'play',
                'stop',
                'eyeglasses',
                'cloud'
            ],
            menu_icon="balloon",
            default_index=0
        )
        if network_menu == "Build rippled":
            form = st.form(key='build_rippled_form')
            form.text('Build rippled using Claudia...')
            repo = form.text_input('Enter the absolute path to the rippled repo', placeholder='e.g. /Users/me/rippled')
            yes_button = form.form_submit_button(label='GO!')
            if yes_button:
                if not os.path.isabs(repo) or not os.path.exists(repo):
                    st.error('Valid absolute repo path must be provided! Please try again.')
                else:
                    with st.spinner(
                            'Building rippled. Please do not navigate away, close the tab or refresh the page '
                            'until the task has finished.'):
                        launch_console_and_execute(get_launch_command_rippled_build(repo))
        elif network_menu == "Install rippled":
            form = st.form(key='install_rippled_form')
            form.text('Install rippled using Claudia...')
            rippled_branch = form.selectbox(
                'Choose rippled branch',
                ('master', 'develop', 'release'),
                index=0
            )

            yes_button = form.form_submit_button(label='GO!')
            if yes_button:
                with st.spinner(
                        'Installing rippled. Please do not navigate away, close the tab or refresh the page '
                        'until the task has finished.'):
                    launch_console_and_execute(get_launch_command_rippled_install(rippled_branch))
        elif network_menu == "Start Network":
            form = st.form(key='start_network_form')
            form.info('Please make sure you have built/installed rippled using Claudia prior to running this step.')
            form.text('Start Local Network using Claudia...')
            yes_button = form.form_submit_button(label='GO!')
            if yes_button:
                with st.spinner(
                        'Starting Local Network. Please do not navigate away, close the tab or refresh the page '
                        'until the task has finished.'):
                    launch_console_and_execute(get_launch_command_network_start())
        elif network_menu == "Stop Network":
            form = st.form(key='stop_network_form')
            form.text('Stop Local Network using Claudia...')
            yes_button = form.form_submit_button(label='GO!')
            if yes_button:
                with st.spinner(
                        'Stopping Local Network. Please do not navigate away, close the tab or refresh the page '
                        'until the task has finished.'):
                    launch_console_and_execute(get_launch_command_network_stop())
        elif network_menu == "Network Status":
            form = st.form(key='check_network_status_form')
            form.info('Please make sure you have built/installed rippled using Claudia prior to running this step.')
            form.text('Check the status of the Local Network using Claudia...')
            yes_button = form.form_submit_button(label='GO!')
            if yes_button:
                with st.spinner(
                        'Checking Local Network Status. Please do not '
                        'navigate away, close the tab or refresh the page until the task has finished.'):
                    launch_console_and_execute(get_launch_command_network_status())
        elif network_menu == "Build Witness Server":
            form = st.form(key='build_witness_server_form')
            form.text('Build Witness Server using Claudia...')
            repo = form.text_input('Enter the absolute path to the Witness Server repo',
                                   placeholder='e.g. /Users/me/xbridge_witness')
            yes_button = form.form_submit_button(label='GO!')
            if yes_button:
                if not os.path.isabs(repo) or not os.path.exists(repo):
                    st.error('Valid absolute repo path must be provided! Please try again.')
                else:
                    with st.spinner(
                            'Building witness server. Please do not navigate away, close the tab or refresh the page '
                            'until the task has finished.'):
                        launch_console_and_execute(get_launch_command_witness_build(repo))
        elif network_menu == "Start Sidechain Network":
            form = st.form(key='start_sidechain_network_form')
            form.info('Before you run this steps, please make sure you have:')
            form.info('   1. built/installed rippled')
            form.info('   2. started local network')
            form.info('   3. built witness server')
            form.text('Start Sidechain Network using Claudia...')
            yes_button = form.form_submit_button(label='GO!')
            if yes_button:
                with st.spinner(
                        'Starting Sidechain Network. Please do not navigate away, close the tab or refresh the page '
                        'until the task has finished.'):
                    launch_console_and_execute(get_launch_command_sidechain_start())
        elif network_menu == "Stop Sidechain Network":
            form = st.form(key='stop_sidechain_network_form')
            form.text('Stop Sidechain Network using Claudia...')
            yes_button = form.form_submit_button(label='GO!')
            if yes_button:
                with st.spinner(
                        'Stopping Sidechain network. Please do not navigate away, close the tab or refresh the page '
                        'until the task has finished.'):
                    launch_console_and_execute(get_launch_command_sidechain_stop())
        elif network_menu == "Check Witness Server Status":
            form = st.form(key='check_witness_server_status')
            form.info('Please make sure you have built the witness server using Claudia prior to running this step.')
            form.text('Check the status of the Witness Server using Claudia...')
            yes_button = form.form_submit_button(label='GO!')
            if yes_button:
                with st.spinner(
                        'Checking Witness Server status. Please do not navigate away, close the tab or refresh the page '
                        'until the task has finished.'):
                    launch_console_and_execute(get_launch_command_witness_status())
        elif network_menu == "Deploy Network to Cloud":
            # form = st.form(key='aws_cloud_keys_form')
            with st.expander('REQUIRED: AWS credentials and rippled build version information', expanded = True):
                col1, col2 = st.columns(2)
                with col1:
                    access_key_id = st.text_input('Enter AWS Access Key ID', type="password")
                    secret_access_key_id = st.text_input('Enter AWS Secret Access Key ID', type="password")
                    region = st.selectbox(
                        'Choose AWS region',
                        ('us-east-1', 'us-east-2', 'us-west-1', 'us-west-2', 'ca-central-1', 'eu-central-1',
                         'ap-south-1'),
                        index=0
                    )
                    check_credentials_button = st.button(label='Check credentials')
                    if check_credentials_button:
                        if are_keys_valid(access_key_id, secret_access_key_id, region):
                            st.success('Keys are valid!')
                        else:
                            st.error('Invalid keys. Please try again')
                with col2:
                    master_node_count = st.slider(
                        'Number of validator nodes running rippled built from master branch',
                        key='master_node_counter',
                        min_value=0,
                        max_value=10,
                        value=0,
                        step=1
                    )
                    release_node_count = st.slider(
                        'Number of validator nodes running rippled built from release branch',
                        key='release_node_counter',
                        min_value=0,
                        max_value=10,
                        value=0,
                        step=1
                    )
                    develop_node_count = st.slider(
                        'Number of validator nodes running rippled built from develop branch',
                        key='develop_node_counter',
                        min_value=0,
                        max_value=10,
                        value=0,
                        step=1
                    )

            yes_button = st.button(label='GO!')
            if yes_button:
                    credentials_valid = are_keys_valid(access_key_id, secret_access_key_id, region)
                    node_count_valid = int(master_node_count) + int(release_node_count) + int(develop_node_count) >= 2

                    if not credentials_valid:
                        st.error('Invalid keys. Please try again.')
                    if not node_count_valid:
                        st.error('At least 2 nodes are required.')
                    if credentials_valid and node_count_valid:
                        with st.spinner(
                                'Deploying Network to Cloud. Please do not navigate away, close the tab or refresh the page '
                                'until the task has finished.'):
                            command = f"python3 aws/aws_helper.py {access_key_id} {secret_access_key_id} {region} {master_node_count} {release_node_count} {develop_node_count}"
                            launch_console_and_execute(command)
elif sidebar_selection == "XRPL Tests":
    with st.container():
        test_menu = option_menu(
            sidebar_selection,
            ["Run System Tests", "Run Unit Tests"],
            icons=['building', "boxes"],
            menu_icon="balloon",
            default_index=0
        )
        if test_menu == "Run Unit Tests":
            form = st.form(key='unit_test_form')
            form.text('Run Unit Tests using Claudia...')
            filter = form.text_input('OPTIONAL: Enter the test filter.', placeholder='e.g. ripple.tx.Offer')
            form.info("If no filter is applied, all unit tests will be executed. This can take a while.")
            yes_button = form.form_submit_button(label='GO!')
            if yes_button:
                with st.spinner(
                        'Running Unit Tests. Please do not navigate away, close the tab or refresh the page '
                        'until the task has finished.'):
                    launch_console_and_execute(get_launch_command_run_unit_tests(filter))
        elif test_menu == "Run System Tests":
            st.text('Run System Tests using Claudia...')
            col1, col2, col3, col4, col5 = st.columns(5)
            with col1:
                lib = st.radio(
                    'Choose client language',
                    ('py', 'js'))
            with col2:
                if lib == "py":
                    client_type = st.radio(
                        'Choose client type',
                        ('jsonrpc', 'websocket'))
                    invalidate_cache = 'false'
                else:
                    client_type = "websocket"
                    invalidate_cache = st.radio(
                        'Invalidate cache?',
                        ('false', 'true'))
            with col3:
                network = st.radio(
                    'Choose network',
                    ('local-mainnet', 'devnet', 'testnet', 'local-sidechain'))
            with col4:
                tag = st.radio(
                    'Choose test tag',
                    ('smoke', 'regression', 'time_intensive'))
            with col5:
                feature = st.radio(
                    'Choose feature',
                    ('payments', 'trustline', 'nft_burn_mint', 'all'))

            yes_button = st.button(label='GO!')
            if yes_button:
                with st.spinner(
                        'Running System Tests. Please do not navigate away, close the tab or refresh the page '
                        'until the task has finished.'):
                    try:
                        explorer_msg = f'Click [here]({get_explorer_url(network)}) to open explorer...'
                        st.markdown(explorer_msg, unsafe_allow_html=True)
                        setup_system_test_env(lib, client_type, network, tag, feature, invalidate_cache)
                        launch_console_and_execute(get_launch_command_system_test_tests(lib, tag, feature))
                        teardown_system_test_env(lib)
                    except Exception as e:
                        set_to_project_root_wd()
                        raise e
elif sidebar_selection == "XRPL Learning Center":
    with st.container():
        col1, col2 = st.columns(2)
        with col1:
            learning_center_menu = option_menu(
                sidebar_selection,
                [
                    "Create an Account",
                    "Send a Payment",
                    "Mint an NFT",
                    "Burn an NFT",
                    "Create an NFT Buy Offer",
                    "Accept an NFT Buy Offer",
                    "Create an NFT Sell Offer",
                    "Accept an NFT Sell Offer"
                ],
                icons=[
                    'book',
                    'book',
                    'book',
                    'book',
                    'book',
                    'book',
                    'book',
                    'book'
                ],
                menu_icon="balloon",
                default_index=0
            )

        with col2:
            if learning_center_menu == "Create an Account":
                create_account()
            elif learning_center_menu == "Send a Payment":
                send_payment_across_accounts()
            elif learning_center_menu == "Mint an NFT":
                mint_nft()
            elif learning_center_menu == "Burn an NFT":
                burn_nft()
            elif learning_center_menu == "Create an NFT Buy Offer":
                create_nft_buy_offer()
            elif learning_center_menu == "Accept an NFT Buy Offer":
                accept_nft_buy_offer()
            elif learning_center_menu == "Create an NFT Sell Offer":
                create_nft_sell_offer()
            elif learning_center_menu == "Accept an NFT Sell Offer":
                accept_nft_sell_offer()
elif sidebar_selection == "Settings":
    with st.container():
        settings_menu = option_menu(
            sidebar_selection,
            [
                "Set Install Mode",
                "Enable a rippled feature",
                "Disable a rippled feature",
                "Cleanup",
                "List System Requirements",
                "List System Test features",
                "Print rippled version"
            ],
            icons=[
                'gear',
                'gear',
                'gear',
                'gear',
                'gear',
                'gear',
                'gear'
            ],
            menu_icon="balloon",
            default_index=0
        )

        if settings_menu == "List System Requirements":
            form = st.form(key='list_system_requirements_form')
            form.text('List System Requirements for running Claudia...')
            yes_button = form.form_submit_button(label='GO!')
            if yes_button:
                with st.spinner(
                        'Listing System Requirements. Please do not navigate away, close the tab or refresh the page '
                        'until the task has finished.'):
                    launch_console_and_display(get_launch_command_system_requirements_message())
        elif settings_menu == "List System Test features":
            form = st.form(key='list_system_tests_features_form')
            form.text('List features available with System Tests...')
            yes_button = form.form_submit_button(label='GO!')
            if yes_button:
                with st.spinner(
                        'Checking all the feature available with System Tests. Please do not navigate away, close the tab or refresh the page '
                        'until the task has finished.'):
                    launch_console_and_display(get_launch_command_system_test_features_message())
        elif settings_menu == "Print rippled version":
            form = st.form(key='print_rippled_version_form')
            form.text('Print rippled version installed/built by Claudia...')
            yes_button = form.form_submit_button(label='GO!')
            if yes_button:
                with st.spinner(
                        'Checking rippled version. Please do not navigate away, close the tab or refresh the page '
                        'until the task has finished.'):
                    launch_console_and_execute(get_launch_command_rippled_version())
        elif settings_menu == "Cleanup":
            form = st.form(key='cleanup_form')
            form.text('Perform system cleanup using Claudia...')
            form.warning(
                "WARNING!! Running this command will wipe out everything built by Claudia. This action cannot be "
                "undone.")
            yes_button = form.form_submit_button(label='GO!')
            if yes_button:
                with st.spinner(
                        'Performing system cleanup. Please do not navigate away, close the tab or refresh the page '
                        'until the task has finished.'):
                    launch_console_and_execute(get_launch_command_clean())
        elif settings_menu == "Enable a rippled feature":
            form = st.form(key='enable_rippled_feature_form')
            form.text('Enable a rippled feature using Claudia...')
            feature = form.text_input('Enter the rippled feature name', placeholder='e.g. DIDoc')
            yes_button = form.form_submit_button(label='GO!')
            if yes_button:
                with st.spinner(
                        f"Attempting to enable '{feature}' feature. Please do not navigate away, close the tab or refresh the page "
                        f"until the task has finished."):
                    launch_console_and_execute(get_launch_command_enable_feature(feature))
        elif settings_menu == "Disable a rippled feature":
            form = st.form(key='disable_rippled_feature_form')
            form.text('Disable a rippled feature using Claudia...')
            feature = form.text_input('Enter the rippled feature name', placeholder='e.g. DIDoc')
            yes_button = form.form_submit_button(label='GO!')
            if yes_button:
                with st.spinner(
                        f"Attempting to disable '{feature}' feature. Please do not navigate away, close the tab or refresh the page "
                        f"until the task has finished."):
                    launch_console_and_execute(get_launch_command_disable_feature(feature))
        elif settings_menu == "Set Install Mode":
            form = st.form(key='set_install_mode_form')
            form.text('Set the install mode to "build" or "install"...')
            mode = form.radio(
                'Choose the install mode',
                ('Build', 'Install'))
            yes_button = form.form_submit_button(label='GO!')
            if yes_button:
                with st.spinner(
                        f"Setting the install mode to {mode}. Please do not navigate away, close the tab or refresh the page "
                        f"until the task has finished."):
                    if mode == "Build":
                        launch_console_and_execute(get_launch_command_set_install_mode_to_build())
                    elif mode == "Install":
                        launch_console_and_execute(get_launch_command_set_install_mode_to_install())



from xrpl import CryptoAlgorithm
from xrpl.account import does_account_exist
from xrpl.clients import JsonRpcClient, WebsocketClient
from xrpl.ledger import get_fee
from xrpl.wallet import Wallet

from claudia.python.lib.exceptions.UnconfiguredEnvironmentException import UnconfiguredEnvironmentException
from claudia.python.lib.framework import constants
from claudia.python.lib.framework.common import send_payment, read_env_var


def before_all(context):

    context.connectionType = read_env_var('CONNECTION_TYPE')
    connectionScheme = read_env_var('CONNECTION_SCHEME')
    connectionURL = read_env_var('CONNECTION_URL')
    context.url = "{}://{}".format(connectionScheme, connectionURL)
    try:
        context.client = get_client(context)
    except ConnectionRefusedError:
        raise Exception("\nERROR: Cannot connect to {}. Make sure the network is accessible.".format(context.url))
    context.algorithm = CryptoAlgorithm.SECP256K1
    context.test_genesis_wallet = get_test_genesis_wallet(context)
    context.default_fee = int(get_fee(context.client))


def get_test_genesis_wallet(context):
    test_genesis_account_id = "rh1HPuRVsYYvThxG2Bs1MfjmrVC73S16Fb"
    test_genesis_account_seed = "snRzwEoNTReyuvz6Fb1CDXcaJUQdp"
    test_genesis_wallet = Wallet.from_seed(seed=test_genesis_account_seed, algorithm=context.algorithm)
    if is_network_local(context):
        if not does_account_exist(test_genesis_account_id, context.client):
            master_account_seed = "snoPBrXtMeMyMHUVTgbuqAfg1SUTb"
            master_genesis_wallet = Wallet.from_seed(seed=master_account_seed, algorithm=context.algorithm)
            send_payment(context, source_wallet=master_genesis_wallet, destination_wallet=test_genesis_wallet,
                         amount=constants.DEFAULT_TEST_GENESIS_ACCOUNT_FUND)
    else:
        if not does_account_exist(test_genesis_account_id, context.client):
            raise UnconfiguredEnvironmentException(
                "Genesis account is not created. Please create and fund the account first.")

    return test_genesis_wallet


def is_network_local(context):
    return "127.0.0" in context.url or "localhost" in context.url


def before_scenario(context, scenario):
    context.minDrop = context.default_fee
    context.transferAmount = "10000"
    context.exception = None
    context.test_status = None
    context.sourceAccountUsername = ""
    context.sourceWallet = None
    context.sourceAddress = ""
    context.sourceBalance = 0
    context.destinationAccountUsername = ""
    context.destinationWallet = None
    context.destinationAddress = ""
    context.destinationBalance = 0


def after_all(context):
    if context.connectionType == "websocket":
        try:
            context.client.close()
        except AttributeError as e:
            pass


def get_client(context):
    if context.connectionType == "jsonrpc":
        client = JsonRpcClient(context.url)
    elif context.connectionType == "websocket":
        client = WebsocketClient(context.url)
        client.open()
    else:
        raise UnconfiguredEnvironmentException("Unsupported CONNECTION_TYPE='{}'".format(context.connectionType))
    return client



import os
import shutil
import sys
import time
from dataclasses import dataclass
from datetime import datetime
from sys import platform

import boto3
import paramiko
from requests import get

IPS = []
PUBLIC_KEYS = []
custom_ec2_instances = []
data_dir = "aws/tmp"
ip_file = f"{data_dir}/ips/network.txt"
validator_port = "2459"
cidr_block = '192.168.1.0/24'
subnet_cidr = '192.168.1.0/24'
destination_cidr_block = '0.0.0.0/0'
vpc_name = 'claudia-new-vpc'
ig_name = 'claudia-new-ig'
rt_name = 'claudia-new-rt'
sn_name = 'claudia-new-sn'
sg_name = 'claudia-new-sg'
instance_type = 't2.micro'
ec2_name = 'claudia-new-infra-instance'
zone = 'a'
key_name = ''
max_rety_attempt = 30


def run_ssh_command(ip_address, command, retry_count=0, print_details=False):
    try:
        if retry_count > max_rety_attempt:
            raise Exception(f"Could not execute command within {max_rety_attempt} attempts.\n"
                            f"Command: {command}\n"
                            f"IP: {ip_address}")
        retry_count += 1
        ssh_client = connect_ssh_client(ip_address)
        stdin, stdout, stderr = ssh_client.exec_command(command)
        output = str(stdout.read())
        error = str(stderr.read())
        if print_details:
            print(f'\nCOMMAND: {command}\n\n')
            print(f'\nSTDOUT:\n{output}\n\n')
            print(f'\nSTDERR:\n{error}\n\n')
        return stdin, output, error
    except Exception as e:
        print(f"SSH connection failed, so we will retry. Error details: {str(e)}\n")
        time.sleep(5)
        return run_ssh_command(ip_address, command, retry_count, print_details)


def sftp_upload_files(ip_address, local_path, remote_path, retry_count=0):
    try:
        if retry_count > max_rety_attempt:
            raise Exception(f"Could not upload files within {max_rety_attempt} attempts.\n"
                            f"Local Path: {local_path}\n"
                            f"Remote Path: {remote_path}\n"
                            f"IP: {ip_address}")
        retry_count += 1
        sftp_client = connect_sftp_client(ip_address)
        sftp_client.put(local_path, remote_path)
        sftp_client.close()
    except Exception as e:
        print(f"SFTP connection failed, so we will retry. Error details: {str(e)}\n")
        time.sleep(5)
        return sftp_upload_files(ip_address, local_path, remote_path, retry_count)


def are_keys_valid(aws_access_key_id, aws_secret_access_key, region_name):
    session = boto3.Session(
        aws_access_key_id=aws_access_key_id,
        aws_secret_access_key=aws_secret_access_key,
        region_name=region_name
    )
    ec2_client = session.client('ec2')

    try:
        ec2_client.describe_instance_status()
        return True
    except Exception as e:
        return False


def get_image_id(region_name):
    if region_name == 'us-east-1':
        image_id = "ami-053b0d53c279acc90"
    elif region_name == 'us-east-2':
        image_id = "ami-024e6efaf93d85776"
    elif region_name == 'us-west-1':
        image_id = "ami-0f8e81a3da6e2510a"
    elif region_name == 'us-west-2':
        image_id = "ami-03f65b8614a860c29"
    elif region_name == 'ca-central-1':
        image_id = "ami-0ea18256de20ecdfc"
    elif region_name == 'eu-central-1':
        image_id = "ami-04e601abe3e1a910f"
    elif region_name == 'ap-south-1':
        image_id = "ami-0f5ee92e2d63afc18"
    else:
        raise Exception(f'{region_name} is not supported')

    return image_id


def get_rippled_branch_list(master_node_count, release_node_count, develop_node_count):
    branches = []
    for _ in range(master_node_count):
        branches.append("master")
    for _ in range(release_node_count):
        branches.append("release")
    for _ in range(develop_node_count):
        branches.append("develop")
    return branches


def create_ec2_instance(aws_access_key_id, aws_secret_access_key, region, master_node_count, release_node_count,
                        develop_node_count):
    rippled_branch_list = get_rippled_branch_list(master_node_count, release_node_count, develop_node_count)
    count = len(rippled_branch_list)

    if os.path.exists(data_dir):
        shutil.rmtree(data_dir)

    global key_name
    session = boto3.Session(
        aws_access_key_id=aws_access_key_id,
        aws_secret_access_key=aws_secret_access_key,
        region_name=region
    )

    client = session.client('ec2', region)
    resource = session.resource('ec2', region)
    ip = get('https://api.ipify.org').text
    vpc = new_create_vpc(client, resource, cidr_block, vpc_name)
    internet_gateway_id = new_create_internet_gateway(client, vpc, ig_name)
    route_table = new_create_route_table(client, vpc, internet_gateway_id, destination_cidr_block, rt_name)
    subnet_id = new_create_subnet(client, vpc, region, zone, route_table, subnet_cidr, sn_name)
    security_group_id = new_create_security_group(client, resource, vpc, ip, sg_name)
    ami = get_image_id(region)
    date_time_stamp = datetime.now().strftime("%Y_%m_%d-%I:%M:%S_%p")
    key_name = f"claudia-key-pair_{date_time_stamp}.pem"
    new_create_key_pair(client, key_name)
    ec2_instances = new_create_ec2_instance(resource, subnet_id, security_group_id, key_name, ami, instance_type, count)
    print("Waiting for the instances to get ready...")
    time.sleep(60)
    os.makedirs(os.path.dirname(ip_file), exist_ok=True)
    f = open(ip_file, "a")

    for index, ec2Instance in enumerate(ec2_instances):
        instance_name = f"{ec2_name}_{str(index + 1)}"
        ec2Instance.create_tags(Tags=[{"Key": "Name", "Value": instance_name}])
        ec2Instance.wait_until_running()
        ec2Instance.reload()
        ip = ec2Instance.public_ip_address
        print(f"Preparing Instance {str(index + 1)}")
        setup_claudia_and_install_rippled(ip, rippled_branch_list[index])
        public_key, validator_token = get_validator_tool_data(ip)
        IPS.append(ec2Instance.private_ip_address)
        PUBLIC_KEYS.append(public_key)
        f.write(f"Public IP: {ec2Instance.public_ip_address}\n")
        f.write(f"Private IP: {ec2Instance.private_ip_address}\n")
        new_instance = EC2Instance(
            index=index,
            region=region,
            id=str(ec2Instance.id),
            name=instance_name,
            type=ec2Instance.instance_type,
            private_ip=ec2Instance.private_ip_address,
            public_ip=ec2Instance.public_ip_address,
            public_dns=ec2Instance.public_dns_name,
            public_key=public_key,
            validator_token=validator_token,
            key_name=key_name,
            rippled_branch=rippled_branch_list[index]
        )
        custom_ec2_instances.append(new_instance)
    f.close()

    for ec2_instance in custom_ec2_instances:
        print(f"Configuring rippled on Instance # {str(ec2_instance.index)}")
        create_local_rippled_config_file(str(ec2_instance.index), IPS, validator_port, ec2_instance.validator_token)
        create_local_validator_file(str(ec2_instance.index), PUBLIC_KEYS)
        configure_rippled(str(ec2_instance.index), ec2_instance.public_ip)

    for ec2_instance in custom_ec2_instances:
        print(f"Starting rippled on Instance # {str(ec2_instance.index)}")
        kill_running_container = 'sudo docker ps -q --filter "name=rippled_server_instance" | ' \
                                 'xargs -r docker stop && docker ps -q --filter "name=rippled_server_instance" | ' \
                                 'xargs -r docker rm'
        start_rippled_command = f"sudo docker run --name rippled_server_instance -d " \
                                f"-p 51234:51234 -p 2459:2459 -p 6006:6006 -p 50051:50051 " \
                                f"-v ./:/opt/ripple/etc rippled_node /opt/ripple/bin/rippled"
        server_info_command = f"sudo docker exec rippled_server_instance /opt/ripple/bin/rippled server_info"
        run_ssh_command(ec2_instance.public_ip, kill_running_container, print_details=False)
        run_ssh_command(ec2_instance.public_ip, start_rippled_command, print_details=False)
        run_ssh_command(ec2_instance.public_ip, server_info_command, print_details=True)

    print(f"Config and validator files are now available at: {os.getcwd()}/aws/tmp/")
    for ec2_instance in custom_ec2_instances:
        ec2_instance.print_info()
        print("Network URLs:")
        print(f" - WebSocket (WS): ws://{ec2_instance.public_ip}:6006")
        print(f" - JSON-RPC (HTTP): http://{ec2_instance.public_ip}:51234")
        print("===")


def setup_claudia_and_install_rippled(ip, rippled_branch):
    commands = [
        'sudo apt-get clean && sudo apt-get update && '
        'sudo apt-get install ca-certificates curl gnupg python3.6 python3-pip -y',

        'sudo curl -fsSL https://deb.nodesource.com/setup_18.x | '
        'sudo -E bash - && sudo apt-get install nodejs npm -y',

        'sudo install -m 0755 -d /etc/apt/keyrings && '
        'curl -fsSL https://download.docker.com/linux/ubuntu/gpg | '
        'sudo gpg --dearmor -o /etc/apt/keyrings/docker.gpg && sudo chmod a+r /etc/apt/keyrings/docker.gpg',

        'echo \
                "deb [arch="$(dpkg --print-architecture)" '
        'signed-by=/etc/apt/keyrings/docker.gpg] https://download.docker.com/linux/ubuntu \
        "$(. /etc/os-release && echo "$VERSION_CODENAME")" stable" | \
        sudo tee /etc/apt/sources.list.d/docker.list > /dev/null',

        'sudo apt-get update',

        'sudo apt-get install docker-ce docker-ce-cli containerd.io docker-buildx-plugin '
        'docker-compose-plugin docker-compose -y',

        'sudo chmod 666 /var/run/docker.sock',

        'sudo docker run hello-world',

        'sudo pip3 install claudia',

        'sudo claudia --version',

        f'claudia rippled install --rippled_branch {rippled_branch}'
    ]

    for command in commands:
        run_ssh_command(ip, command, print_details=False)


def get_validator_tool_data(ip, print_sensitive_data=False):
    command = 'sudo docker run rippled_node /bin/bash -c "/opt/ripple/bin/validator-keys create_keys ' \
              '--keyfile keys.json > /dev/null && /opt/ripple/bin/validator-keys create_token --keyfile keys.json"'
    stdin, output, error = run_ssh_command(ip, command)
    validator_token = output.split('[validator_token]')[1].replace("\\n", "").replace("'", "")
    public_key = output.split('validator public key: ')[1].split('[validator_token]')[0] \
        .replace("\\n", "").replace("'", "")
    if print_sensitive_data:
        print("\n\n\nGETTING GET_VALIDATOR_TOOL_DATA")
        print(f"validator_token: {validator_token}")
        print(f"public_key: {public_key}")
    return public_key, validator_token


def create_local_rippled_config_file(instance_index, ips, port, token):
    src = "aws/rippled-example.cfg"
    dst = f"aws/tmp/configs/rippled_{instance_index}/rippled.cfg"
    if os.path.exists(dst):
        os.remove(dst)

    os.makedirs(os.path.dirname(dst), exist_ok=True)
    shutil.copy(src, dst)
    f = open(dst, "a")
    f.write("\n[ips_fixed]\n")
    for ip in ips:
        f.write(f"{ip} {port}\n")

    f.write(f"\n[validator_token]\n{token}\n\n")
    f.close()


def create_local_validator_file(instance_index, keys):
    dst = f"aws/tmp/configs/rippled_{instance_index}/validators.txt"
    os.makedirs(os.path.dirname(dst), exist_ok=True)
    f = open(dst, "w")
    f.write("[validators]\n")
    for key in keys:
        f.write(f"{key}\n")

    f.close()


def get_ssh_client():
    ssh = paramiko.SSHClient()
    ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
    return ssh


def connect_ssh_client(ip, username='ubuntu'):
    ssh_client = get_ssh_client()
    privkey = paramiko.RSAKey.from_private_key_file(key_name)
    ssh_client.connect(hostname=ip, username=username, pkey=privkey, timeout=10)
    return ssh_client


def connect_sftp_client(ip, username='ubuntu'):
    ssh_client = connect_ssh_client(ip, username)
    sftp_client = ssh_client.open_sftp()
    return sftp_client


def configure_rippled(instance_index, ip):
    local_rippled_config_file_path = f"{os.getcwd()}/aws/tmp/configs/rippled_{instance_index}/rippled.cfg"
    local_validator_file_path = f"{os.getcwd()}/aws/tmp/configs/rippled_{instance_index}/validators.txt"
    remote_rippled_config_file_path = 'rippled.cfg'
    remote_validator_file_path = 'validators.txt'
    sftp_upload_files(ip, local_rippled_config_file_path, remote_rippled_config_file_path)
    sftp_upload_files(ip, local_validator_file_path, remote_validator_file_path)


def new_create_vpc(client, resource, cidr_block, vpc_name):
    existing_vpcs = client.describe_vpcs(
        Filters=[
            {
                'Name': 'tag:Name',
                'Values': [
                    vpc_name,
                ]
            }
        ]
    )['Vpcs']
    vpc_count = len(existing_vpcs)
    if vpc_count == 0:
        vpcInit = client.create_vpc(CidrBlock=cidr_block)
        vpc = resource.Vpc(vpcInit["Vpc"]["VpcId"])
        vpc.create_tags(Tags=[{"Key": "Name", "Value": vpc_name}])
        vpc.wait_until_available()
        return vpc
    elif vpc_count == 1:
        existing_vpc = resource.Vpc(existing_vpcs[0]["VpcId"])
        return existing_vpc
    else:
        raise Exception(
            f"Multiple VPCs with name: {vpc_name} found. Please delete these before proceeding.")


def new_create_internet_gateway(client, vpc, ig_name):
    existing_internet_gateways = client.describe_internet_gateways(
        Filters=[
            {
                'Name': 'tag:Name',
                'Values': [
                    ig_name,
                ]
            }
        ]
    )['InternetGateways']
    ig_count = len(existing_internet_gateways)

    if ig_count == 0:
        ig_init = client.create_internet_gateway(
            TagSpecifications=[
                {'ResourceType': 'internet-gateway',
                 'Tags': [{"Key": "Name", "Value": ig_name}]}, ]
        )
        ig_id = ig_init["InternetGateway"]["InternetGatewayId"]
        vpc.attach_internet_gateway(InternetGatewayId=ig_id)
        return ig_id
    elif ig_count == 1:
        existing_ig_id = existing_internet_gateways[0]['InternetGatewayId']
        return existing_ig_id
    else:
        raise Exception(
            f"Multiple internet gateways with name: {ig_name} found. Please delete these before proceeding.")


def new_create_route_table(client, vpc, ig_id, destination_cidr_block, rt_name):
    existing_route_tables = client.describe_route_tables(
        Filters=[
            {
                'Name': 'tag:Name',
                'Values': [
                    rt_name,
                ]
            }
        ]
    )['RouteTables']
    route_table_count = len(existing_route_tables)

    if route_table_count == 0:
        routeTable = vpc.create_route_table()
        routeTable.create_route(
            DestinationCidrBlock=destination_cidr_block,
            GatewayId=ig_id
        )
        routeTable.create_tags(Tags=[{"Key": "Name", "Value": rt_name}])
        return routeTable
    elif route_table_count == 1:

        existing_route_table = existing_route_tables[0]
        return existing_route_table
    else:
        raise Exception(f"Multiple route tables with name: {rt_name} found. Please delete these before proceeding.")


def new_create_subnet(client, vpc, region, zone, route_table, subnet_cidr, sn_name):
    existing_subnets = client.describe_subnets(
        Filters=[
            {
                'Name': 'tag:Name',
                'Values': [
                    sn_name,
                ]
            }
        ]
    )['Subnets']
    subnet_count = len(existing_subnets)

    if subnet_count == 0:
        subnet = vpc.create_subnet(
            CidrBlock=subnet_cidr, AvailabilityZone="{}{}".format(region, zone))
        subnet.create_tags(Tags=[{"Key": "Name", "Value": sn_name}])
        route_table.associate_with_subnet(SubnetId=subnet.id)
        return subnet.id
    elif subnet_count == 1:
        existing_subnet_id = existing_subnets[0]['SubnetId']
        return existing_subnet_id
    else:
        raise Exception(f"Multiple subnets with name: {sn_name} found. Please delete these before proceeding.")


def new_create_security_group(client, resource, vpc, host_ip, sg_name):
    existing_security_groups = client.describe_security_groups(
        Filters=[
            {
                'Name': 'group-name',
                'Values': [
                    sg_name,
                ]
            }
        ]
    )['SecurityGroups']
    sg_count = len(existing_security_groups)
    if sg_count == 0:
        secGroup = resource.create_security_group(
            GroupName=sg_name, Description='Claudia Security Group', VpcId=vpc.id)

        secGroup.authorize_ingress(
            IpPermissions=[
                {
                    'FromPort': 22,
                    'ToPort': 22,
                    'IpProtocol': 'tcp',
                    'IpRanges': [
                        {
                            'CidrIp': '{}/32'.format(host_ip),
                            'Description': 'Host SSH Access'
                        },
                    ],
                    'UserIdGroupPairs': [
                        {
                            'GroupId': secGroup.id,
                            'Description': 'Security Group SSH Access'
                        }
                    ],
                },
                {
                    "FromPort": -1,
                    "ToPort": -1,
                    "IpProtocol": "icmp",
                    'IpRanges': [
                        {
                            'CidrIp': '{}/32'.format(host_ip),
                            'Description': 'Host Ping Access'
                        }
                    ],
                    'UserIdGroupPairs': [
                        {
                            'GroupId': secGroup.id,
                            'Description': 'Security Group Ping Access'
                        }
                    ],
                },
                {
                    "FromPort": 80,
                    'ToPort': 80,
                    "IpProtocol": "tcp",
                    'IpRanges': [
                        {
                            'CidrIp': '{}/32'.format(host_ip),
                            'Description': 'Host HTTP Access'
                        }
                    ],
                    'UserIdGroupPairs': [
                        {
                            'GroupId': secGroup.id,
                            'Description': 'Security Group HTTP Access'
                        }
                    ],
                },
                {
                    "FromPort": 443,
                    'ToPort': 443,
                    "IpProtocol": "tcp",
                    'IpRanges': [
                        {
                            'CidrIp': '{}/32'.format(host_ip),
                            'Description': 'Host HTTPS Access'
                        }
                    ],
                    'UserIdGroupPairs': [
                        {
                            'GroupId': secGroup.id,
                            'Description': 'Security Group HTTPS Access'
                        }
                    ],
                },
                {
                    "FromPort": 51234,
                    'ToPort': 51234,
                    "IpProtocol": "tcp",
                    'IpRanges': [
                        {
                            'CidrIp': '{}/32'.format(host_ip),
                            'Description': 'Host 51234 Access'
                        }
                    ],
                    'UserIdGroupPairs': [
                        {
                            'GroupId': secGroup.id,
                            'Description': 'Security Group 51234 Access'
                        }
                    ],
                },
                {
                    "FromPort": 2459,
                    'ToPort': 2459,
                    "IpProtocol": "tcp",
                    'IpRanges': [
                        {
                            'CidrIp': '{}/32'.format(host_ip),
                            'Description': 'Host 2459 Access'
                        }
                    ],
                    'UserIdGroupPairs': [
                        {
                            'GroupId': secGroup.id,
                            'Description': 'Security Group 2459 Access'
                        }
                    ],
                },
                {
                    "FromPort": 6006,
                    'ToPort': 6006,
                    "IpProtocol": "tcp",
                    'IpRanges': [
                        {
                            'CidrIp': '{}/32'.format(host_ip),
                            'Description': 'Host 6006 Access'
                        }
                    ],
                    'UserIdGroupPairs': [
                        {
                            'GroupId': secGroup.id,
                            'Description': 'Security Group 6006 Access'
                        }
                    ],
                },
                {
                    "FromPort": 50051,
                    'ToPort': 50051,
                    "IpProtocol": "tcp",
                    'IpRanges': [
                        {
                            'CidrIp': '{}/32'.format(host_ip),
                            'Description': 'Host 50051 Access'
                        }
                    ],
                    'UserIdGroupPairs': [
                        {
                            'GroupId': secGroup.id,
                            'Description': 'Security Group 50051 Access'
                        }
                    ],
                }
            ]
        )

        secGroup.create_tags(Tags=[{"Key": "Name", "Value": sg_name}])
        return secGroup.group_id
    elif sg_count == 1:
        existing_sg_id = existing_security_groups[0]['GroupId']
        return existing_sg_id
    else:
        raise Exception(f"Multiple security groups with name: {sg_name} found. Please delete these before proceeding.")


def new_create_key_pair(client, key_file_name):
    keyPair = client.create_key_pair(KeyName=key_file_name)
    if os.path.exists(key_file_name):
        os.remove(key_file_name)
    privateKeyFile = open(key_file_name, "w")
    privateKeyFile.write(dict(keyPair)['KeyMaterial'])
    privateKeyFile.close()
    if platform == "linux" or platform == "linux2" or platform == "darwin":
        os.chmod(key_file_name, 0o400)


def new_create_ec2_instance(resource, subnet_id, security_group_id, key_file_name, ami, instance_type, count):
    ec2_instances = resource.create_instances(
        ImageId=ami, InstanceType=instance_type, MaxCount=count, MinCount=count,
        NetworkInterfaces=[{'SubnetId': subnet_id, 'DeviceIndex': 0,
                            'AssociatePublicIpAddress': True, 'Groups': [security_group_id]}],
        KeyName=key_file_name)

    print(f"Created {count} EC2 instances successfully.")
    return ec2_instances


if __name__ == '__main__':
    aws_access_key_id = str(sys.argv[1])
    aws_secret_access_key = str(sys.argv[2])
    region = str(sys.argv[3])
    master_node_count = int(sys.argv[4])
    release_node_count = int(sys.argv[5])
    develop_node_count = int(sys.argv[6])
    print(f"Using these settings to deploy the network:\n"
          f"\taws_access_key_id: {aws_access_key_id}\n"
          f"\taws_secret_access_key: {aws_secret_access_key}\n"
          f"\tregion: {region}\n"
          f"\tmaster_node_count: {master_node_count}\n"
          f"\trelease_node_count: {release_node_count}\n"
          f"\tdevelop_node_count: {develop_node_count}\n")

    create_ec2_instance(
        aws_access_key_id,
        aws_secret_access_key,
        region,
        master_node_count,
        release_node_count,
        develop_node_count)


@dataclass
class EC2Instance:
    index: int
    region: str
    id: str
    name: str
    type: str
    private_ip: str
    public_ip: str
    public_dns: str
    public_key: str
    validator_token: str
    key_name: str
    rippled_branch: str

    def print_info(self):
        print(f'##########################')
        print(f'### Instance {self.index + 1} Details:###')
        print(f'##########################')
        print(f"\t- Instance Id: {self.id}")
        print(f"\t- Instance Name: {self.name}")
        print(f"\t- Private IP Address: {self.private_ip}")
        print(f"\t- Public DNS Name: {self.public_dns}")
        print(f"\t- Public IP Address: {self.public_ip}")
        print(f"\t- Public Key: {self.public_key}")
        print(f"\t- Public Validator Token: {self.validator_token}")
        print(f"\t- Rippled Source Branch: {self.rippled_branch}")
        print(f"\t- SSH connection command: \"cd {os.getcwd()};"
              f"ssh -i '{self.key_name}' ubuntu@{self.public_ip}\"\n")



import streamlit as st
from claudia.claudia import set_python_launch_vars
from claudia.python.features.environment import get_client, get_test_genesis_wallet
from claudia.python.lib.framework import constants
from claudia.python.lib.framework.common import send_payment, ResponseCode, get_account_balance, read_env_var, \
    sign_autofill_and_submit, get_nft_tokens, get_token_offers_response
from claudia.python.lib.framework.object_factory import ObjFactory, ObjType
from xrpl import CryptoAlgorithm
from xrpl.ledger import get_fee
from xrpl.wallet import Wallet


class Context(object):
    connectionType = None
    url = None
    client = None
    algorithm = None
    test_genesis_wallet = None
    default_fee = None
    minDrop = None
    transferAmount = None
    exception = None
    test_status = None


def init_context():
    try:
        context = Context()
        context.connectionType = read_env_var('CONNECTION_TYPE')
        connectionScheme = read_env_var('CONNECTION_SCHEME')
        connectionURL = read_env_var('CONNECTION_URL')
        context.url = "{}://{}".format(connectionScheme, connectionURL)
        try:
            context.client = get_client(context)
        except ConnectionRefusedError:
            raise Exception(f"ERROR: Cannot connect to {context.url}. Make sure the network is accessible.")
        context.algorithm = CryptoAlgorithm.SECP256K1
        context.test_genesis_wallet = get_test_genesis_wallet(context)
        context.default_fee = int(get_fee(context.client))
        context.minDrop = context.default_fee
        context.transferAmount = constants.DEFAULT_TRANSFER_AMOUNT
        context.exception = None
        context.test_status = None
        return context
    except Exception as e:
        st.error(f"An unexpected error has occurred. Details: \n{e}")
        raise e


def verify_network(context):
    if not context.client.is_open():
        st.error("Network is not accessible!")
        raise Exception("Network is not accessible!")


def create_wallet(context):
    try:
        wallet = Wallet.create(algorithm=context.algorithm)
        return wallet
    except Exception as e:
        return f"An unexpected error has occurred. Details: \n{e}"


def fund_wallet(context, wallet, fund_amount=constants.DEFAULT_ACCOUNT_FUND_AMOUNT):
    try:
        response = send_payment(
            context,
            source_wallet=context.test_genesis_wallet,
            destination_wallet=wallet,
            amount=fund_amount
        )
        return response
    except Exception as e:
        return f"An unexpected error has occurred. Details: \n{e}"


def verify_funding(response):
    try:
        if ResponseCode.success not in str(response):
            raise Exception(
                "Expected response code was not received."
                "\nExpected Code = '{}'."
                "\nResponse received: '{}'".format(ResponseCode.success, response)
            )
    except Exception as e:
        return f"An unexpected error has occurred. Details: \n{e}"


def get_create_account_summary(context, wallet):
    return f"- Workflow Summary:" \
           f"\n   - Wallet Address: {wallet.classic_address}" \
           f"\n   - Wallet Seed: {wallet.seed}" \
           f"\n   - Wallet Funds: {str(get_account_balance(context, wallet.classic_address))}" \
           f"\n   - Source Wallet Address: {context.test_genesis_wallet.classic_address}"


def get_send_payment_summary(context, source_wallet, destination_wallet, initial_source_balance,
                             initial_destination_balance):
    return f"- Workflow Summary:" \
           f"\n   - Source Wallet Address: {source_wallet.classic_address}" \
           f"\n   - Source Wallet Seed: {source_wallet.seed}" \
           f"\n   - Source Wallet Initial Balance: {initial_source_balance}" \
           f"\n   - Source Wallet Current Balance: {str(get_account_balance(context, source_wallet.classic_address))}" \
           f"\n   - Destination Wallet Address: {destination_wallet.classic_address}" \
           f"\n   - Destination Wallet Seed: {destination_wallet.seed}" \
           f"\n   - Destination Wallet Initial Balance: {initial_destination_balance}" \
           f"\n   - Destination Wallet Current Balance: {str(get_account_balance(context, destination_wallet.classic_address))}"


def get_nft_minting_summary(context, owner_address, owner_seed, nftoken):
    return f" - Workflow Summary:" \
           f"\n    - Owner Wallet Address: {owner_address}" \
           f"\n    - Owner Wallet Seed: {owner_seed}" \
           f"\n    - Minted NFT ID: {nftoken}"


def get_nft_burning_summary(owner_address,
                            owner_seed,
                            nft_id,
                            initial_owner_nft_count,
                            owner_nft_count_post_burning):
    return f"- Workflow Summary:" \
           f"\n   - Owner Wallet Address: {owner_address}" \
           f"\n   - Owner Wallet Seed: {owner_seed}" \
           f"\n   - Burnt NFT ID: {nft_id}" \
           f"\n   - Owner Wallet NFT Count before burning: {initial_owner_nft_count}" \
           f"\n   - Owner Wallet NFT Count after burning: {owner_nft_count_post_burning}"


def get_create_buy_offer_summary(
        owner_address,
        owner_seed,
        buyer_address,
        buyer_seed,
        nft_id,
        offer_id
):
    return f"- Workflow Summary:" \
           f"\n   - Owner Wallet Address: {owner_address}" \
           f"\n   - Owner Wallet Seed: {owner_seed}" \
           f"\n   - Buyer Wallet Address: {buyer_address}" \
           f"\n   - Buyer Wallet Seed: {buyer_seed}" \
           f"\n   - Minted NFT ID: {nft_id}" \
           f"\n   - Created Offer ID: {offer_id}"


def get_accept_buy_offer_summary(
        owner_address,
        owner_seed,
        offer_id
):
    return f"- Workflow Summary:" \
           f"\n   - Owner Wallet Address: {owner_address}" \
           f"\n   - Owner Wallet Seed: {owner_seed}" \
           f"\n   - Accepted Buy Offer ID: {offer_id}"


def get_create_sell_offer_summary(
        owner_address,
        owner_seed,
        nft_id,
        offer_id
):
    return f"- Workflow Summary:" \
           f"\n   - Owner Wallet Address: {owner_address}" \
           f"\n   - Owner Wallet Seed: {owner_seed}" \
           f"\n   - Minted NFT ID: {nft_id}" \
           f"\n   - Created Offer ID: {offer_id}"


def get_accept_sell_offer_summary(
        buyer_address,
        buyer_seed,
        offer_id
):
    return f"- Workflow Summary:" \
           f"\n   - Buyer Wallet Address: {buyer_address}" \
           f"\n   - Buyer Wallet Seed: {buyer_seed}" \
           f"\n   - Accepted Sell Offer ID: {offer_id}"


def propose_or_get_wallet(context, seed=""):
    if seed == "":
        wallet = Wallet.create(algorithm=context.algorithm)
        send_payment(context,
                     source_wallet=context.test_genesis_wallet,
                     destination_wallet=wallet,
                     amount=constants.DEFAULT_ACCOUNT_FUND_AMOUNT)
    else:
        wallet = Wallet.from_seed(seed=seed, algorithm=context.algorithm)

    return wallet, wallet.seed


def ask_for_network():
    network = st.selectbox(
        'Choose Network',
        ('local-mainnet', 'devnet', 'testnet', 'local-sidechain'))
    yes_button = st.button(label='GO!')
    if yes_button:
        set_python_launch_vars(network, "websocket")
        return


def validate_input(key, value):
    if value == "" or value is None:
        error = f"{key} is a required field. Please try again..."
        st.error(error)
        raise Exception(error)


def create_account():
    network = st.selectbox(
        'Choose Network',
        ('local-mainnet', 'devnet', 'testnet', 'local-sidechain'))
    continue_button = st.button(label='Go!', use_container_width=True)
    if continue_button:
        with st.spinner(''):
            st.caption("Details")
            with st.container():
                st.text(" - Applying settings and verifying if the network is accessible")
                set_python_launch_vars(network, "websocket")
                context = init_context()
                verify_network(context)
                st.text(" - Proposing wallet")
                wallet = create_wallet(context)
                st.text(" - Funding wallet")
                response = fund_wallet(context, wallet)
                st.text(" - Verifying wallet funding")
                verify_funding(response)
                st.text('- Account created! 🎉')
                st.code(get_create_account_summary(context, wallet))


def send_payment_across_accounts():
    network = st.selectbox(
        'Choose Network',
        ('local-mainnet', 'devnet', 'testnet', 'local-sidechain'))

    with st.expander(" - OPTIONAL: Enter source wallet seed. If nothing is provided, Claudia will create one"):
        source_seed = st.text_input('Source Wallet Seed', type="password", key='source_seed_send_payment')

    with st.expander(" - OPTIONAL: Enter destination wallet seed. If nothing is provided, Claudia will create one"):
        destination_seed = st.text_input('Destination Wallet Seed', type="password",
                                         key='destination_seed_send_payment')

    with st.expander(" - OPTIONAL: Enter fund amount. If nothing is provided, Claudia will send 1000 drops"):
        fund_amount = st.text_input('Fund Amount in XRP drops', "1000", key='fund_send_payment')
    continue_button = st.button(label='Go!', use_container_width=True)
    if continue_button:
        with st.spinner(''):
            st.caption("Details")
            with st.container():
                st.text(" - Applying settings and verifying if the network is accessible")
                set_python_launch_vars(network, "websocket")
                context = init_context()
                verify_network(context)
                st.text(" - Getting the wallets ready")
                source_wallet, source_seed = propose_or_get_wallet(context, source_seed)
                destination_wallet, destination_seed = propose_or_get_wallet(context, destination_seed)
                initial_source_balance = get_account_balance(context, source_wallet.classic_address)
                initial_destination_balance = get_account_balance(context, source_wallet.classic_address)
                st.text(" - Sending payment")
                response = send_payment(
                    context,
                    source_wallet=source_wallet,
                    destination_wallet=destination_wallet,
                    amount=fund_amount
                )
                st.text(" - Verifying payment")
                verify_funding(response)
                st.text('- Payment was successful! 🎉')
                st.code(get_send_payment_summary(context,
                                                 source_wallet,
                                                 destination_wallet,
                                                 initial_source_balance,
                                                 initial_destination_balance), language='bash')


def mint_nft():
    network = st.selectbox(
        'Choose Network',
        ('local-mainnet', 'devnet', 'testnet', 'local-sidechain'))

    with st.expander(" - OPTIONAL: Enter owner wallet seed. If nothing is provided, Claudia will create one"):
        owner_seed = st.text_input('Owner Wallet Seed', type="password", key='owner_seed_mint')
    continue_button = st.button(label='Go!', use_container_width=True)
    if continue_button:
        with st.spinner(''):
            st.caption("Details")
            with st.container():
                st.text(" - Applying settings and verifying if the network is accessible")
                set_python_launch_vars(network, "websocket")
                context = init_context()
                verify_network(context)
                st.text(" - Getting the owner wallet ready")
                owner_wallet, owner_seed = propose_or_get_wallet(context, owner_seed)
                owner_address = owner_wallet.classic_address
                owner_seed = owner_wallet.seed
                st.text(" - Minting NFT")
                payload = ObjFactory.getObj(
                    ObjType.nft_token_mint,
                    account=owner_address,
                    nftoken_taxon=0,
                    flags=8
                )
                response = sign_autofill_and_submit(context.client, payload, owner_wallet)
                st.text(f" - Verifying if minting succeeded")
                if ResponseCode.success not in str(response):
                    raise Exception(
                        "Expected response code was not received."
                        "\nExpected Code = '{}'."
                        "\nResponse received: '{}'".format(ResponseCode.success, response)
                    )

                nftoken = get_nft_tokens(context.client, owner_address)[-1]
                st.text('- NFT Minting was successful! 🎉')
                st.code(
                    get_nft_minting_summary(
                        context,
                        owner_address,
                        owner_seed,
                        nftoken
                    ),
                    language='bash'
                )


def burn_nft():
    network = st.selectbox(
        'Choose Network',
        ('local-mainnet', 'devnet', 'testnet', 'local-sidechain'))

    owner_seed = st.text_input('REQUIRED: Owner Wallet Seed', type="password", key='owner_seed_burn')
    nft_id = st.text_input('REQUIRED: NFT ID', key='nft_burn')
    continue_button = st.button(label='Go!', use_container_width=True)
    if continue_button:
        validate_input("Owner Seed", owner_seed)
        validate_input("NFT ID", nft_id)
        with st.spinner(''):
            st.caption("Details")
            with st.container():
                st.text(" - Applying settings and verifying if the network is accessible")
                set_python_launch_vars(network, "websocket")
                context = init_context()
                verify_network(context)
                st.text(" - Getting the owner wallet ready")
                owner_wallet, owner_seed = propose_or_get_wallet(context, owner_seed)
                owner_address = owner_wallet.classic_address
                initial_owner_nft_count = len(get_nft_tokens(context.client, owner_address))
                st.text(" - Burning NFT")
                payload = ObjFactory.getObj(
                    ObjType.nft_token_burn,
                    account=owner_address,
                    nftoken_id=nft_id,
                )
                response = sign_autofill_and_submit(context.client, payload, owner_wallet)
                owner_nft_count_post_burning = len(get_nft_tokens(context.client, owner_address))

                st.text(f" - Verifying if burning succeeded")
                if ResponseCode.success not in str(response):
                    raise Exception(
                        "Expected response code was not received."
                        "\nExpected Code = '{}'."
                        "\nResponse received: '{}'".format(ResponseCode.success, response)
                    )

                st.text('- NFT Minting was successful! 🎉')
                st.code(
                    get_nft_burning_summary(
                        owner_address,
                        owner_seed,
                        nft_id,
                        initial_owner_nft_count,
                        owner_nft_count_post_burning
                    ),
                    language='bash'
                )


def create_nft_buy_offer():
    network = st.selectbox(
        'Choose Network',
        ('local-mainnet', 'devnet', 'testnet', 'local-sidechain'))

    owner_seed = st.text_input('REQUIRED: Owner Wallet Seed', type="password", key='owner_seed_create_buy')
    nft_id = st.text_input('REQUIRED: NFT ID', key='nft_create_buy')

    with st.expander(" - OPTIONAL: Enter buyer wallet seed. If nothing is provided, Claudia will create one"):
        buyer_seed = st.text_input('Buyer Wallet Seed', type="password", key='buyer_seed_create_buy')
    continue_button = st.button(label='Go!', use_container_width=True)
    if continue_button:
        validate_input("Owner Seed", owner_seed)
        validate_input("NFT ID", nft_id)
        with st.spinner(''):
            st.caption("Details")
            with st.container():
                st.text(" - Applying settings and verifying if the network is accessible")
                set_python_launch_vars(network, "websocket")
                context = init_context()
                verify_network(context)
                st.text(" - Getting the wallets ready")
                owner_wallet, owner_seed = propose_or_get_wallet(context, owner_seed)
                owner_address = owner_wallet.classic_address
                buyer_wallet, buyer_seed = propose_or_get_wallet(context, buyer_seed)
                buyer_address = buyer_wallet.classic_address
                st.text(f" - Creating a buy offer")
                payload = ObjFactory.getObj(
                    ObjType.nf_token_create_offer,
                    account=buyer_address,
                    owner=owner_address,
                    nftoken_id=nft_id,
                    amount=context.transferAmount
                )
                response = sign_autofill_and_submit(context.client, payload, buyer_wallet)

                st.text(f" - Verifying if offer creation succeeded")
                if ResponseCode.success not in str(response):
                    raise Exception(
                        "Expected response code was not received."
                        "\nExpected Code = '{}'."
                        "\nResponse received: '{}'".format(ResponseCode.success, response)
                    )

                offer_id = get_token_offers_response(context.client, buyer_address, token_id=nft_id)[0]
                st.text('- Creating NFT Buy Offer was successful! 🎉')
                st.code(
                    get_create_buy_offer_summary(
                        owner_address,
                        owner_seed,
                        buyer_address,
                        buyer_seed,
                        nft_id,
                        offer_id
                    ),
                    language='bash'
                )


def accept_nft_buy_offer():
    network = st.selectbox(
        'Choose Network',
        ('local-mainnet', 'devnet', 'testnet', 'local-sidechain'))

    owner_seed = st.text_input('REQUIRED: Owner Wallet Seed', type="password", key='owner_seed_accept_buy')
    offer_id = st.text_input('REQUIRED: Buy Offer ID', key='offer_accept_buy')

    continue_button = st.button(label='Go!', use_container_width=True)
    if continue_button:
        validate_input("Owner Seed", owner_seed)
        validate_input("Buy Offer ID", offer_id)
        with st.spinner(''):
            st.caption("Details")
            with st.container():
                st.text(" - Applying settings and verifying if the network is accessible")
                set_python_launch_vars(network, "websocket")
                context = init_context()
                verify_network(context)
                st.text(" - Getting the wallets ready")
                owner_wallet, owner_seed = propose_or_get_wallet(context, owner_seed)
                owner_address = owner_wallet.classic_address
                st.text(f" - Accepting a buy offer")
                payload = ObjFactory.getObj(
                    ObjType.nf_token_accept_offer,
                    account=owner_address,
                    nftoken_buy_offer=offer_id
                )
                response = sign_autofill_and_submit(context.client, payload, owner_wallet)

                st.text(f" - Verifying if offer creation succeeded")
                if ResponseCode.success not in str(response):
                    raise Exception(
                        "Expected response code was not received."
                        "\nExpected Code = '{}'."
                        "\nResponse received: '{}'".format(ResponseCode.success, response)
                    )
                st.text('- Creating NFT Buy Offer was successful! 🎉')
                st.code(
                    get_accept_buy_offer_summary(
                        owner_address,
                        owner_seed,
                        offer_id
                    ),
                    language='bash'
                )


def create_nft_sell_offer():
    network = st.selectbox(
        'Choose Network',
        ('local-mainnet', 'devnet', 'testnet', 'local-sidechain'))

    owner_seed = st.text_input('REQUIRED: Owner Wallet Seed', type="password", key='owner_seed_create_sell')
    nft_id = st.text_input('REQUIRED: NFT ID', key='nft_create_sell')
    continue_button = st.button(label='Go!', use_container_width=True)
    if continue_button:
        validate_input("Owner Seed", owner_seed)
        validate_input("NFT ID", nft_id)
        with st.spinner(''):
            st.caption("Details")
            with st.container():
                st.text(" - Applying settings and verifying if the network is accessible")
                set_python_launch_vars(network, "websocket")
                context = init_context()
                verify_network(context)
                st.text(" - Getting the wallets ready")
                owner_wallet, owner_seed = propose_or_get_wallet(context, owner_seed)
                owner_address = owner_wallet.classic_address
                st.text(f" - Creating a sell offer")
                payload = ObjFactory.getObj(
                    ObjType.nf_token_create_offer,
                    account=owner_address,
                    nftoken_id=nft_id,
                    amount=context.transferAmount,
                    flags=1
                )
                response = sign_autofill_and_submit(context.client, payload, owner_wallet)

                st.text(f" - Verifying if offer creation succeeded")
                if ResponseCode.success not in str(response):
                    raise Exception(
                        "Expected response code was not received."
                        "\nExpected Code = '{}'."
                        "\nResponse received: '{}'".format(ResponseCode.success, response)
                    )

                offer_id = get_token_offers_response(context.client, owner_address, token_id=nft_id)[0]

                st.text('- Creating NFT Sell Offer was successful! 🎉')
                st.code(
                    get_create_sell_offer_summary(
                        owner_address,
                        owner_seed,
                        nft_id,
                        offer_id
                    ),
                    language='bash'
                )


def accept_nft_sell_offer():
    network = st.selectbox(
        'Choose Network',
        ('local-mainnet', 'devnet', 'testnet', 'local-sidechain'))

    offer_id = st.text_input('REQUIRED: Sell Offer ID', key='offer_accept_sell')
    with st.expander(" - OPTIONAL: Enter buyer wallet seed. If nothing is provided, Claudia will create one"):
        buyer_seed = st.text_input('Buyer Wallet Seed', type="password", key='buyer_seed_accept_sell')

    continue_button = st.button(label='Go!', use_container_width=True)
    if continue_button:
        validate_input("Sell Offer ID", offer_id)
        with st.spinner(''):
            st.caption("Details")
            with st.container():
                st.text(" - Applying settings and verifying if the network is accessible")
                set_python_launch_vars(network, "websocket")
                context = init_context()
                verify_network(context)
                st.text(" - Getting the wallets ready")
                buyer_wallet, buyer_seed = propose_or_get_wallet(context, buyer_seed)
                buyer_address = buyer_wallet.classic_address
                st.text(f" - Accepting a sell offer")
                payload = ObjFactory.getObj(
                    ObjType.nf_token_accept_offer,
                    account=buyer_address,
                    nftoken_sell_offer=offer_id
                )
                response = sign_autofill_and_submit(context.client, payload, buyer_wallet)

                st.text(f" - Verifying if offer creation succeeded")
                if ResponseCode.success not in str(response):
                    raise Exception(
                        "Expected response code was not received."
                        "\nExpected Code = '{}'."
                        "\nResponse received: '{}'".format(ResponseCode.success, response)
                    )
                st.text('- Accepting NFT Sell Offer was successful! 🎉')
                st.code(
                    get_accept_sell_offer_summary(
                        buyer_address,
                        buyer_seed,
                        offer_id
                    ),
                    language='bash'
                )
